#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from typing import Union

import numpy as np
import torch
from torch.utils.data import Dataset

from disent.dataset.data import GroundTruthData
from disent.dataset.wrapper._base import WrappedDataset
from disent.util.math.random import random_choice_prng

log = logging.getLogger(__name__)


# ========================================================================= #
# Masked Dataset                                                            #
# ========================================================================= #


DataTypeHint = Union[GroundTruthData, np.ndarray, torch.Tensor]
MaskTypeHint = Union[str, np.ndarray]


def load_mask_indices(length: int, mask_or_indices: MaskTypeHint):
    # load as numpy mask if it is a string!
    if isinstance(mask_or_indices, str):
        mask_or_indices = np.load(mask_or_indices)["mask"]
    # load data
    assert isinstance(mask_or_indices, np.ndarray)
    # check inputs
    if mask_or_indices.dtype == "bool":
        # boolean values
        assert length == len(mask_or_indices)
        indices = np.arange(length)[mask_or_indices]
    else:
        # integer values
        assert len(np.unique(mask_or_indices)) == len(mask_or_indices)
        assert np.min(mask_or_indices) >= 0
        assert np.max(mask_or_indices) < length
        indices = mask_or_indices
    # check that we have at least 1 value
    assert len(indices) > 0
    assert len(indices) <= length
    # return values
    return indices


class MaskedDataset(WrappedDataset):
    def __init__(self, data: DataTypeHint, mask: MaskTypeHint, randomize: bool = False):
        assert isinstance(data, (GroundTruthData, torch.Tensor, np.ndarray))
        n = len(data)
        # save values
        self._data = data
        self._indices = load_mask_indices(n, mask)
        # randomize
        if randomize:
            l = len(self._indices)
            self._indices = load_mask_indices(n, random_choice_prng(n, size=l, replace=False))
            assert len(self._indices) == l
            log.info(f"replaced mask: {l}/{n} ({l/n:.3f}) with randomized mask!")

    def __len__(self):
        return len(self._indices)

    def __getitem__(self, idx):
        return self._data[self._indices[idx]]

    @property
    def data(self) -> Dataset:
        return self._data


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


# ========================================================================= #
# Numba Is An Optional Dependency                                           #
# ========================================================================= #


def try_njit(*args, **kwargs):
    """
    Wrapper around numba.njit
    - If numba is installed, then we JIT the decorated function
    - If numba is missing, then we do nothing and leave the function untouched!
    """
    try:
        from numba import njit
    except ImportError:
        # dummy njit
        def njit(*args, **kwargs):
            def _wrapper(func):
                import warnings

                warnings.warn(f"failed to JIT compile: {func}, numba is not installed!")
                return func

            return _wrapper

    # try and JIT compile function!
    return njit(*args, **kwargs)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from fractions import Fraction
from typing import List
from typing import Optional
from typing import Union

import numpy as np

from disent.dataset.data import GroundTruthData
from disent.dataset.sampling._base import BaseDisentSampler
from disent.dataset.util.state_space import StateSpace

# ========================================================================= #
# Ground Truth Dist Sampler                                                 #
# ========================================================================= #


class GroundTruthDistSampler(BaseDisentSampler):
    def uninit_copy(self) -> "GroundTruthDistSampler":
        return GroundTruthDistSampler(
            num_samples=self._num_samples,
            triplet_sample_mode=self._triplet_sample_mode,
            triplet_swap_chance=self._triplet_swap_chance,
        )

    def __init__(
        self,
        num_samples=1,
        triplet_sample_mode="manhattan_scaled",
        triplet_swap_chance=0.0,
    ):
        super().__init__(num_samples=num_samples)
        # checks
        assert num_samples in {1, 2, 3}, f"num_samples ({repr(num_samples)}) must be 1, 2 or 3"
        assert triplet_sample_mode in {
            "random",
            "factors",
            "manhattan",
            "manhattan_scaled",
            "combined",
            "combined_scaled",
        }, f'sample_mode ({repr(triplet_sample_mode)}) must be one of {["random", "factors", "manhattan", "combined"]}'
        # save hparams
        self._num_samples = num_samples
        self._triplet_sample_mode = triplet_sample_mode
        self._triplet_swap_chance = triplet_swap_chance
        # scaled
        self._scaled = False
        if triplet_sample_mode.endswith("_scaled"):
            triplet_sample_mode = triplet_sample_mode[: -len("_scaled")]
            self._scaled = True
        # checks
        assert triplet_sample_mode in {"random", "factors", "manhattan", "combined"}, "It is a bug if this fails!"
        assert 0 <= triplet_swap_chance <= 1, "triplet_swap_chance must be in range [0, 1]"
        # set vars
        self._sample_mode = triplet_sample_mode
        self._swap_chance = triplet_swap_chance
        # dataset variable
        self._state_space: Optional[StateSpace] = None

    def _init(self, dataset):
        assert isinstance(
            dataset, GroundTruthData
        ), f"dataset must be an instance of {repr(GroundTruthData.__class__.__name__)}, got: {repr(dataset)}"
        self._state_space = dataset.state_space_copy()

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Sampling                                                              #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _sample_idx(self, idx):
        # sample indices
        indices = (idx, *np.random.randint(0, len(self._state_space), size=self._num_samples - 1))
        # sort based on mode
        if self._num_samples == 3:
            a_i, p_i, n_i = self._swap_triple(indices)
            # randomly swap positive and negative
            if np.random.random() < self._swap_chance:
                indices = (a_i, n_i, p_i)
            else:
                indices = (a_i, p_i, n_i)
        # get data
        return indices

    def _swap_triple(self, indices):
        a_i, p_i, n_i = indices
        a_f, p_f, n_f = self._state_space.idx_to_pos(indices)
        # get the scale for everything
        # - range of positions is [0, f_size - 1], to scale between 0 and 1 we need to
        #   divide by (f_size - 1), but if the factor size is 1, we can't divide by zero
        #   so we make the minimum 1
        scale = np.maximum(1, self._state_space.factor_sizes - 1) if (self._scaled) else None
        # SWAP: manhattan
        if self._sample_mode == "manhattan":
            if factor_dist(a_f, p_f, scale=scale) > factor_dist(a_f, n_f, scale=scale):
                return a_i, n_i, p_i
        # SWAP: factors
        elif self._sample_mode == "factors":
            if factor_diff(a_f, p_f) > factor_diff(a_f, n_f):
                return a_i, n_i, p_i
        # SWAP: combined
        elif self._sample_mode == "combined":
            if factor_diff(a_f, p_f) > factor_diff(a_f, n_f):
                return a_i, n_i, p_i
            elif factor_diff(a_f, p_f) == factor_diff(a_f, n_f):
                if factor_dist(a_f, p_f, scale=scale) > factor_dist(a_f, n_f, scale=scale):
                    return a_i, n_i, p_i
        # SWAP: random
        elif self._sample_mode != "random":
            raise KeyError("invalid mode")
        # done!
        return indices


def factor_diff(f0: np.ndarray, f1: np.ndarray) -> int:
    # input types should be np.int64
    assert f0.dtype == f1.dtype == "int64"
    # compute distances!
    return np.sum(f0 != f1)


# NOTE: scaling here should always be the same as `disentangle_loss`
def factor_dist(f0: np.ndarray, f1: np.ndarray, scale: np.ndarray = None) -> Union[Fraction, int]:
    # compute distances!
    if scale is None:
        # input types should all be np.int64
        assert f0.dtype == f1.dtype == "int64", f"invalid dtypes, f0: {f0.dtype}, f1: {f1.dtype}"
        # we can simply sum if everything is already an integer
        return np.sum(np.abs(f0 - f1))
    else:
        # input types should all be np.int64
        assert f0.dtype == f1.dtype == scale.dtype == "int64"
        # Division results in precision errors! We cannot simply sum divided values. We instead
        # store values as arbitrary precision rational numbers in the form of fractions This means
        # we do not lose precision while summing, and avoid comparison errors!
        #    - https://shlegeris.com/2018/10/23/sqrt.html
        #    - https://cstheory.stackexchange.com/a/4010
        # 1. first we need to convert numbers to python arbitrary precision values:
        f0: List[int] = f0.tolist()
        f1: List[int] = f1.tolist()
        scale: List[int] = scale.tolist()
        # 2. we need to sum values in the form of fractions
        total = Fraction(0)
        for y0, y1, s in zip(f0, f1, scale):
            total += Fraction(abs(y0 - y1), s)
        return total


# ========================================================================= #
# Investigation:                                                            #
# ========================================================================= #


if __name__ == "__main__":

    def main():
        from tqdm import tqdm

        from disent.dataset import DisentDataset
        from disent.dataset.data import Cars3d64Data
        from disent.dataset.data import DSpritesData
        from disent.dataset.data import Shapes3dData
        from disent.dataset.data import SmallNorb64Data
        from disent.dataset.data import XYObjectData
        from disent.dataset.data import XYObjectShadedData
        from disent.dataset.data import XYSquaresMinimalData
        from disent.util.seeds import TempNumpySeed

        repeats = 1000
        samples = 100

        # RESULTS - manhattan:
        #   cars3d:             orig_vs_divs=30.066%, orig_vs_frac=30.066%, divs_vs_frac=0.000%
        #   3dshapes:           orig_vs_divs=12.902%, orig_vs_frac=12.878%, divs_vs_frac=0.096%
        #   dsprites:           orig_vs_divs=24.035%, orig_vs_frac=24.032%, divs_vs_frac=0.003%
        #   smallnorb:          orig_vs_divs=18.601%, orig_vs_frac=18.598%, divs_vs_frac=0.005%
        #   xy_squares_minimal: orig_vs_divs= 1.389%, orig_vs_frac= 0.000%, divs_vs_frac=1.389%
        #   xy_object:          orig_vs_divs=15.520%, orig_vs_frac=15.511%, divs_vs_frac=0.029%
        #   xy_object:          orig_vs_divs=23.973%, orig_vs_frac=23.957%, divs_vs_frac=0.082%
        # RESULTS - combined:
        #   cars3d:             orig_vs_divs=15.428%, orig_vs_frac=15.428%, divs_vs_frac=0.000%
        #   3dshapes:           orig_vs_divs=4.982%,  orig_vs_frac= 4.968%, divs_vs_frac=0.050%
        #   dsprites:           orig_vs_divs=8.366%,  orig_vs_frac= 8.363%, divs_vs_frac=0.003%
        #   smallnorb:          orig_vs_divs=7.359%,  orig_vs_frac= 7.359%, divs_vs_frac=0.000%
        #   xy_squares_minimal: orig_vs_divs=0.610%,  orig_vs_frac= 0.000%, divs_vs_frac=0.610%
        #   xy_object:          orig_vs_divs=7.622%,  orig_vs_frac= 7.614%, divs_vs_frac=0.020%
        #   xy_object:          orig_vs_divs=8.741%,  orig_vs_frac= 8.733%, divs_vs_frac=0.046%
        for mode in ["manhattan", "combined"]:
            for data_cls in [
                Cars3d64Data,
                Shapes3dData,
                DSpritesData,
                SmallNorb64Data,
                XYSquaresMinimalData,
                XYObjectData,
                XYObjectShadedData,
            ]:
                data = data_cls()
                dataset_orig = DisentDataset(data, sampler=GroundTruthDistSampler(3, f"{mode}"))
                dataset_frac = DisentDataset(data, sampler=GroundTruthDistSampler(3, f"{mode}_scaled"))
                dataset_divs = DisentDataset(data, sampler=GroundTruthDistSampler(3, f"{mode}_scaled_INVALID"))
                # calculate the average number of mismatches between sampling methods!
                all_wrong_frac = []  # frac vs orig
                all_wrong_divs = []  # divs vs orig
                all_wrong_diff = []  # frac vs divs
                with TempNumpySeed(777):
                    progress = tqdm(range(repeats), desc=f"{mode} {data.name}")
                    for i in progress:
                        batch_seed = np.random.randint(0, 2**32)
                        with TempNumpySeed(batch_seed):
                            idxs_orig = np.array(
                                [
                                    dataset_orig.sampler.sample(np.random.randint(0, len(dataset_orig)))
                                    for _ in range(samples)
                                ]
                            )
                        with TempNumpySeed(batch_seed):
                            idxs_frac = np.array(
                                [
                                    dataset_frac.sampler.sample(np.random.randint(0, len(dataset_frac)))
                                    for _ in range(samples)
                                ]
                            )
                        with TempNumpySeed(batch_seed):
                            idxs_divs = np.array(
                                [
                                    dataset_divs.sampler.sample(np.random.randint(0, len(dataset_divs)))
                                    for _ in range(samples)
                                ]
                            )
                        # check number of miss_matches
                        all_wrong_frac.append(np.sum(np.any(idxs_orig != idxs_frac, axis=-1)) / samples * 100)
                        all_wrong_divs.append(np.sum(np.any(idxs_orig != idxs_divs, axis=-1)) / samples * 100)
                        all_wrong_diff.append(np.sum(np.any(idxs_frac != idxs_divs, axis=-1)) / samples * 100)
                        # update progress bar
                        progress.set_postfix(
                            {
                                "orig_vs_divs": f"{np.mean(all_wrong_divs):5.3f}%",
                                "orig_vs_frac": f"{np.mean(all_wrong_frac):5.3f}%",
                                "divs_vs_frac": f"{np.mean(all_wrong_diff):5.3f}%",
                            }
                        )

    main()


# ========================================================================= #
# END:                                                                      #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import numpy as np
import pytest
import torch
from scipy.stats import gmean
from scipy.stats import hmean

from disent.dataset import DisentDataset
from disent.dataset.data import XYObjectData
from disent.dataset.sampling import RandomSampler
from disent.dataset.transform import ToImgTensorF32
from disent.nn.functional import torch_conv2d_channel_wise
from disent.nn.functional import torch_conv2d_channel_wise_fft
from disent.nn.functional import torch_corr_matrix
from disent.nn.functional import torch_cov_matrix
from disent.nn.functional import torch_dct
from disent.nn.functional import torch_dct2
from disent.nn.functional import torch_dist
from disent.nn.functional import torch_dist_hamming
from disent.nn.functional import torch_gaussian_kernel_2d
from disent.nn.functional import torch_idct
from disent.nn.functional import torch_idct2
from disent.nn.functional import torch_mean_generalized
from disent.nn.functional import torch_norm
from disent.nn.functional import torch_norm_euclidean
from disent.nn.functional import torch_norm_manhattan
from disent.util import to_numpy

# ========================================================================= #
# TESTS                                                                     #
# ========================================================================= #


def test_cov_corr():
    for i in range(5, 1000, 250):
        for j in range(2, 100, 25):
            # these match when torch.float64 is used, not when torch float32 is used...
            xs = torch.randn(i, j, dtype=torch.float64)

            np_cov = torch.from_numpy(np.cov(to_numpy(xs), rowvar=False, ddof=0)).to(xs.dtype)
            np_cor = torch.from_numpy(np.corrcoef(to_numpy(xs), rowvar=False, ddof=0)).to(xs.dtype)

            cov = torch_cov_matrix(xs)
            cor = torch_corr_matrix(xs)

            assert torch.allclose(np_cov, cov)
            assert torch.allclose(np_cor, cor)


def test_generalised_mean():
    xs = torch.abs(torch.randn(2, 1000, 3, dtype=torch.float64))

    # normal
    assert torch.allclose(torch_mean_generalized(xs, p="arithmetic", dim=1), torch.mean(xs, dim=1))
    assert torch.allclose(torch_mean_generalized(xs, p=1, dim=1), torch.mean(xs, dim=1))

    # scipy equivalents
    assert torch.allclose(torch_mean_generalized(xs, p="geometric", dim=1), torch.as_tensor(gmean(xs, axis=1)))
    assert torch.allclose(torch_mean_generalized(xs, p="harmonic", dim=1), torch.as_tensor(hmean(xs, axis=1)))
    assert torch.allclose(torch_mean_generalized(xs, p=0, dim=1), torch.as_tensor(gmean(xs, axis=1)))
    assert torch.allclose(torch_mean_generalized(xs, p=-1, dim=1), torch.as_tensor(hmean(xs, axis=1)))
    assert torch.allclose(
        torch_mean_generalized(xs, p=0), torch.as_tensor(gmean(xs, axis=None))
    )  # scipy default axis is 0
    assert torch.allclose(
        torch_mean_generalized(xs, p=-1), torch.as_tensor(hmean(xs, axis=None))
    )  # scipy default axis is 0

    # min max
    assert torch.allclose(torch_mean_generalized(xs, p="maximum", dim=1), torch.amax(xs, dim=1))
    assert torch.allclose(torch_mean_generalized(xs, p="minimum", dim=1), torch.amin(xs, dim=1))
    assert torch.allclose(torch_mean_generalized(xs, p=np.inf, dim=1), torch.amax(xs, dim=1))
    assert torch.allclose(torch_mean_generalized(xs, p=-np.inf, dim=1), torch.amin(xs, dim=1))


def test_p_norm():
    xs = torch.abs(torch.randn(2, 1000, 3, dtype=torch.float64))

    inf = float("inf")

    # torch equivalents
    assert torch.allclose(torch_norm(xs, p=1, dim=-1), torch.linalg.norm(xs, ord=1, dim=-1))
    assert torch.allclose(torch_norm(xs, p=2, dim=-1), torch.linalg.norm(xs, ord=2, dim=-1))
    assert torch.allclose(torch_norm(xs, p="maximum", dim=-1), torch.linalg.norm(xs, ord=inf, dim=-1))

    # torch equivalents -- less than zero [FAIL]
    with pytest.raises(ValueError, match="p-norm cannot have a p value less than 1"):
        assert torch.allclose(torch_norm(xs, p=0, dim=-1), torch.linalg.norm(xs, ord=0, dim=-1))
    with pytest.raises(ValueError, match="p-norm cannot have a p value less than 1"):
        assert torch.allclose(torch_norm(xs, p="minimum", dim=-1), torch.linalg.norm(xs, ord=-inf, dim=-1))

    # torch equivalents -- less than zero
    assert torch.allclose(torch_dist(xs, p=0, dim=-1), torch.linalg.norm(xs, ord=0, dim=-1))
    assert torch.allclose(torch_dist(xs, p="minimum", dim=-1), torch.linalg.norm(xs, ord=-inf, dim=-1))
    assert torch.allclose(torch_dist(xs, p=0, dim=-1), torch.linalg.norm(xs, ord=0, dim=-1))
    assert torch.allclose(torch_dist(xs, p="minimum", dim=-1), torch.linalg.norm(xs, ord=-inf, dim=-1))

    # test other axes
    ys = torch.flatten(torch.moveaxis(xs, 0, -1), start_dim=1, end_dim=-1)
    assert torch.allclose(torch_norm(xs, p=2, dim=[0, -1]), torch.linalg.norm(ys, ord=2, dim=-1))
    ys = torch.flatten(xs, start_dim=1, end_dim=-1)
    assert torch.allclose(torch_norm(xs, p=1, dim=[-2, -1]), torch.linalg.norm(ys, ord=1, dim=-1))
    ys = torch.flatten(torch.moveaxis(xs, -1, 0), start_dim=1, end_dim=-1)
    assert torch.allclose(torch_dist(xs, p=0, dim=[0, 1]), torch.linalg.norm(ys, ord=0, dim=-1))

    # check equal names
    assert torch.allclose(torch_dist(xs, dim=1, p="euclidean"), torch_norm_euclidean(xs, dim=1))
    assert torch.allclose(torch_dist(xs, dim=1, p=2), torch_norm_euclidean(xs, dim=1))
    assert torch.allclose(torch_dist(xs, dim=1, p="manhattan"), torch_norm_manhattan(xs, dim=1))
    assert torch.allclose(torch_dist(xs, dim=1, p=1), torch_norm_manhattan(xs, dim=1))
    assert torch.allclose(torch_dist(xs, dim=1, p="hamming"), torch_dist_hamming(xs, dim=1))
    assert torch.allclose(torch_dist(xs, dim=1, p=0), torch_dist_hamming(xs, dim=1))

    # check axes
    assert torch_dist(xs, dim=1, p=2, keepdim=False).shape == (2, 3)
    assert torch_dist(xs, dim=1, p=2, keepdim=True).shape == (2, 1, 3)
    assert torch_dist(xs, dim=-1, p=1, keepdim=False).shape == (2, 1000)
    assert torch_dist(xs, dim=-1, p=1, keepdim=True).shape == (2, 1000, 1)
    assert torch_dist(xs, dim=0, p=0, keepdim=False).shape == (1000, 3)
    assert torch_dist(xs, dim=0, p=0, keepdim=True).shape == (1, 1000, 3)
    assert torch_dist(xs, dim=[0, -1], p=-inf, keepdim=False).shape == (1000,)
    assert torch_dist(xs, dim=[0, -1], p=-inf, keepdim=True).shape == (1, 1000, 1)
    assert torch_dist(xs, dim=[0, 1], p=inf, keepdim=False).shape == (3,)
    assert torch_dist(xs, dim=[0, 1], p=inf, keepdim=True).shape == (1, 1, 3)

    # check norm over all
    assert torch_dist(xs, dim=None, p=0, keepdim=False).shape == ()
    assert torch_dist(xs, dim=None, p=0, keepdim=True).shape == (1, 1, 1)
    assert torch_dist(xs, dim=None, p=1, keepdim=False).shape == ()
    assert torch_dist(xs, dim=None, p=1, keepdim=True).shape == (1, 1, 1)
    assert torch_dist(xs, dim=None, p=2, keepdim=False).shape == ()
    assert torch_dist(xs, dim=None, p=2, keepdim=True).shape == (1, 1, 1)
    assert torch_dist(xs, dim=None, p=inf, keepdim=False).shape == ()
    assert torch_dist(xs, dim=None, p=inf, keepdim=True).shape == (1, 1, 1)


def test_dct():
    x = torch.randn(128, 3, 64, 32, dtype=torch.float64)

    # chceck +ve dims
    assert torch.allclose(x, torch_idct(torch_dct(x, dim=0), dim=0))
    with pytest.raises(ValueError, match="does not support odd sized dimension"):
        torch.allclose(x, torch_idct(torch_dct(x, dim=1), dim=1))
    assert torch.allclose(x, torch_idct(torch_dct(x, dim=2), dim=2))
    assert torch.allclose(x, torch_idct(torch_dct(x, dim=3), dim=3))

    # chceck -ve dims
    assert torch.allclose(x, torch_idct(torch_dct(x, dim=-4), dim=-4))
    with pytest.raises(ValueError, match="does not support odd sized dimension"):
        torch.allclose(x, torch_idct(torch_dct(x, dim=-3), dim=-3))
    assert torch.allclose(x, torch_idct(torch_dct(x, dim=-2), dim=-2))
    assert torch.allclose(x, torch_idct(torch_dct(x, dim=-1), dim=-1))

    # check defaults
    assert torch.allclose(torch_dct(x), torch_dct(x, dim=-1))
    assert torch.allclose(torch_idct(x), torch_idct(x, dim=-1))

    # check dct2
    assert torch.allclose(x, torch_idct2(torch_dct2(x)))
    assert torch.allclose(x, torch_idct2(torch_dct2(x)))

    # check defaults dct2
    assert torch.allclose(torch_dct2(x), torch_dct2(x, dim1=-1, dim2=-2))
    assert torch.allclose(torch_dct2(x), torch_dct2(x, dim1=-2, dim2=-1))
    assert torch.allclose(torch_idct2(x), torch_idct2(x, dim1=-1, dim2=-2))
    assert torch.allclose(torch_idct2(x), torch_idct2(x, dim1=-2, dim2=-1))
    # check order dct2
    assert torch.allclose(torch_dct2(x, dim1=-1, dim2=-2), torch_dct2(x, dim1=-2, dim2=-1))
    assert torch.allclose(torch_dct2(x, dim1=-1, dim2=-4), torch_dct2(x, dim1=-4, dim2=-1))
    assert torch.allclose(torch_dct2(x, dim1=-4, dim2=-1), torch_dct2(x, dim1=-1, dim2=-4))
    assert torch.allclose(torch_idct2(x, dim1=-1, dim2=-2), torch_idct2(x, dim1=-2, dim2=-1))
    assert torch.allclose(torch_idct2(x, dim1=-1, dim2=-4), torch_idct2(x, dim1=-4, dim2=-1))
    assert torch.allclose(torch_idct2(x, dim1=-4, dim2=-1), torch_idct2(x, dim1=-1, dim2=-4))


def test_fft_conv2d():
    data = XYObjectData()
    dataset = DisentDataset(data, RandomSampler(), transform=ToImgTensorF32(), augment=None)
    # sample data
    factors = dataset.gt_data.sample_random_factor_traversal(f_idx=2)
    batch = dataset.dataset_batch_from_factors(factors=factors, mode="input")
    # test torch_conv2d_channel_wise variants
    for i in range(1, 5):
        kernel = torch_gaussian_kernel_2d(sigma=i)
        out_cnv = torch_conv2d_channel_wise(signal=batch, kernel=kernel)[0]
        out_fft = torch_conv2d_channel_wise_fft(signal=batch, kernel=kernel)[0]
        assert torch.amax(torch.abs(out_cnv - out_fft)) < 1e-6


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from typing import Optional

from disent.dataset.data import GroundTruthData
from disent.dataset.sampling._base import BaseDisentSampler
from disent.dataset.util.state_space import StateSpace

log = logging.getLogger(__name__)


# ========================================================================= #
# Convert ground truth data to a dataset                                    #
# ========================================================================= #


class GroundTruthSingleSampler(BaseDisentSampler):
    def uninit_copy(self) -> "GroundTruthSingleSampler":
        return GroundTruthSingleSampler()

    def __init__(self):
        super().__init__(num_samples=1)
        self._state_space: Optional[StateSpace] = None  # TODO: not actually needed

    def _init(self, dataset):
        assert isinstance(
            dataset, GroundTruthData
        ), f"dataset must be an instance of {repr(GroundTruthData.__class__.__name__)}, got: {repr(dataset)}"
        self._state_space = dataset.state_space_copy()

    def _sample_idx(self, idx):
        return (idx,)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging

import numpy as np
from torch.utils.data import Dataset

from disent.dataset.data import GroundTruthData
from disent.dataset.util.state_space import StateSpace
from disent.dataset.wrapper._base import WrappedDataset
from disent.util.math.dither import nd_dither_matrix

log = logging.getLogger(__name__)


# ========================================================================= #
# Dithered Dataset                                                          #
# ========================================================================= #


class DitheredDataset(WrappedDataset):
    def __init__(self, gt_data: GroundTruthData, dither_n: int = 2, keep_ratio: float = 1):
        assert 0 < keep_ratio <= 1.0
        assert isinstance(gt_data, GroundTruthData)
        # -~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
        self._gt_data = gt_data
        # data space
        data_idx = np.arange(len(gt_data))
        data_pos = gt_data.idx_to_pos(data_idx)
        # dmat space
        d_mat = nd_dither_matrix(n=dither_n, d=self._gt_data.num_factors, norm=True) < keep_ratio
        d_states = StateSpace(d_mat.shape)
        # data space to dmat space
        dmat_pos = data_pos % dither_n
        dmat_idx = d_states.pos_to_idx(dmat_pos)
        mask = d_mat.flatten()[dmat_idx]
        # convert mask to indices
        self._indices = np.arange(len(gt_data))[mask]
        # -~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
        assert len(self._indices) > 0
        # check values & count compared to ratio
        unique_count_map = {False: 0, True: 0, **{u: c for u, c in zip(*np.unique(mask, return_counts=True))}}
        assert len(unique_count_map) == 2
        assert unique_count_map[True] > 0
        assert sum(unique_count_map.values()) == len(gt_data)
        log.info(
            f"[n={dither_n}] keep ratio: {keep_ratio:.2f} actual ratio: {unique_count_map[True] / sum(unique_count_map.values()):.2f}"
        )

    def __len__(self):
        return len(self._indices)

    def __getitem__(self, item):
        return self._gt_data[self._indices[item]]

    @property
    def data(self) -> Dataset:
        return self._gt_data


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np

from disent.dataset.data import GroundTruthData
from disent.dataset.sampling._base import BaseDisentSampler
from disent.dataset.util.state_space import StateSpace
from disent.util.math.random import sample_radius

log = logging.getLogger(__name__)


# ========================================================================= #
# triplets                                                                  #
# ========================================================================= #


class GroundTruthTripleSampler(BaseDisentSampler):
    def uninit_copy(self) -> "GroundTruthTripleSampler":
        return GroundTruthTripleSampler(
            p_k_range=self.p_k_range,
            n_k_range=self.n_k_range,
            n_k_sample_mode=self.n_k_sample_mode,
            n_k_is_shared=self.n_k_is_shared,
            p_radius_range=self.p_radius_range,
            n_radius_range=self.n_radius_range,
            n_radius_sample_mode=self.n_radius_sample_mode,
            swap_metric=self._swap_metric,
            swap_chance=self._swap_chance,
        )

    def __init__(
        self,
        # factor sampling
        p_k_range=(0, -1),
        n_k_range=(0, -1),
        n_k_sample_mode="bounded_below",
        n_k_is_shared=True,
        # radius sampling
        p_radius_range=(0, -1),
        n_radius_range=(0, -1),
        n_radius_sample_mode="bounded_below",
        # final checks
        swap_metric=None,
        swap_chance=None,
    ):
        super().__init__(num_samples=3)
        # checks
        assert swap_metric in {
            None,
            "k",
            "manhattan",
            "manhattan_norm",
            "euclidean",
            "euclidean_norm",
        }, f"Invalid {swap_metric=}"
        assert n_k_sample_mode in {"offset", "bounded_below", "random"}, f"Invalid {n_k_sample_mode=}"
        assert n_radius_sample_mode in {"offset", "bounded_below", "random"}, f"Invalid {n_radius_sample_mode=}"
        # factors
        self.p_k_range = p_k_range
        self.n_k_range = n_k_range
        self.n_k_sample_mode = n_k_sample_mode
        self.n_k_is_shared = n_k_is_shared
        # radius
        self.p_radius_range = p_radius_range
        self.n_radius_range = n_radius_range
        self.n_radius_sample_mode = n_radius_sample_mode
        # SWAP: if negative is not further than the positive
        self._swap_metric = swap_metric
        self._swap_chance = swap_chance
        if swap_chance is not None:
            assert 0 <= swap_chance <= 1, f"{swap_chance=} must be in range 0 to 1."
        # dataset variable
        self._state_space: Optional[StateSpace]

    def _init(self, dataset):
        assert isinstance(
            dataset, GroundTruthData
        ), f"dataset must be an instance of {repr(GroundTruthData.__class__.__name__)}, got: {repr(dataset)}"
        self._state_space = dataset.state_space_copy()
        # DIFFERING FACTORS
        self.p_k_min, self.p_k_max, self.n_k_min, self.n_k_max = self._min_max_from_range(
            p_range=self.p_k_range,
            n_range=self.n_k_range,
            max_values=self._state_space.num_factors,
            n_sample_mode=self.n_k_sample_mode,
            is_radius=False,
        )
        # RADIUS SAMPLING
        self.p_radius_min, self.p_radius_max, self.n_radius_min, self.n_radius_max = self._min_max_from_range(
            p_range=self.p_radius_range,
            n_range=self.n_radius_range,
            max_values=self._state_space.factor_sizes,
            n_sample_mode=self.n_radius_sample_mode,
            is_radius=True,
        )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # CORE                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _sample_idx(self, idx):
        f0, f1, f2 = self.datapoint_sample_factors_triplet(idx)
        return (
            self._state_space.pos_to_idx(f0),
            self._state_space.pos_to_idx(f1),
            self._state_space.pos_to_idx(f2),
        )

    def datapoint_sample_factors_triplet(self, idx):
        # SAMPLE FACTOR INDICES
        p_k, n_k = self._sample_num_factors()
        p_shared_indices, n_shared_indices = self._sample_shared_indices(p_k, n_k)
        # SAMPLE FACTORS - sample, resample and replace shared factors with originals
        anchor_factors = self._state_space.idx_to_pos(idx)
        positive_factors, negative_factors = self._resample_factors(anchor_factors)
        positive_factors[p_shared_indices] = anchor_factors[p_shared_indices]
        negative_factors[n_shared_indices] = anchor_factors[n_shared_indices]
        # SWAP IF +VE FURTHER THAN -VE
        if self._swap_metric is not None:
            positive_factors, negative_factors = self._swap_factors(anchor_factors, positive_factors, negative_factors)
        # RANDOMLY SWAP +ve AND -ve IF CHANCE:
        if self._swap_chance is not None:
            if np.random.random() < self._swap_chance:
                positive_factors, negative_factors = negative_factors, positive_factors
        # return factors!
        return anchor_factors, positive_factors, negative_factors

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # HELPER                                                                #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _min_max_from_range(self, p_range, n_range, max_values, n_sample_mode, is_radius=False):
        p_min, p_max = normalise_range_pair(p_range, max_values)
        n_min, n_max = normalise_range_pair(n_range, max_values)
        # cross factor assertions
        if n_sample_mode == "offset":
            if is_radius:
                # TODO: radius for -ve values should be handled when sampling, not here.
                if not np.all((p_max + n_min) <= np.floor_divide(max_values, 2)):
                    raise FactorSizeError(
                        f"Factor dimensions are too small for given offset radius:"
                        f"\n\tUnsatisfied: p_max + offset_min <= max_size // 2"
                        f"\n\tUnsatisfied: {p_max} + {n_min} <= {np.array(max_values)} // 2"
                        f"\n\tUnsatisfied: {p_max + n_min} <= {np.floor_divide(max_values, 2)}"
                    )
            else:
                if not np.all((p_max + n_min) <= max_values):
                    raise FactorSizeError(
                        f"Factor dimensions are too small for given offset range:"
                        f"\n\tUnsatisfied: p_max + offset_min <= max_size"
                        f"\n\tUnsatisfied: {p_max} + {n_min} <= {np.array(max_values)}"
                    )
        elif n_sample_mode == "bounded_below":
            if not (np.all(p_max <= n_max)):
                raise FactorSizeError(
                    f"Ranges are not staggered."
                    f"\n\tUnsatisfied: p_max <= n_max"
                    f"\n\tUnsatisfied: {p_max} <= {n_max}"
                )
            if not (np.all(p_max <= max_values) and np.all(n_max <= max_values)):
                raise FactorSizeError(
                    "Factor dimensions are too small for given range:"
                    f"\n\tUnsatisfied: p_max <= max_size and n_max <= max_size"
                    f"\n\tUnsatisfied: {p_max} <= {np.array(max_values)} and {n_max} <= {np.array(max_values)}"
                )
        else:
            if not (np.all(p_min <= n_min) and np.all(p_max <= n_max)):
                raise FactorSizeError(
                    f"Ranges are not staggered. Should be p <= n:"
                    f"\n\tUnsatisfied: p_min <= n_min and p_max <= n_max"
                    f"\n\tUnsatisfied: {p_min} <= {n_min} and {p_max} <= {n_max}"
                )
            if not (np.all(p_max <= max_values) and np.all(n_max <= max_values)):
                raise FactorSizeError(
                    "Factor dimensions are too small for given range:"
                    f"\n\tUnsatisfied: p_max <= max_size and n_max <= max_size"
                    f"\n\tUnsatisfied: {p_max} <= {np.array(max_values)} and {n_max} <= {np.array(max_values)}"
                )
        # we're done!
        return p_min, p_max, n_min, n_max

    def _sample_num_factors(self):
        p_k = np.random.randint(self.p_k_min, self.p_k_max + 1)
        # sample for negative
        if self.n_k_sample_mode == "offset":
            n_k = np.random.randint(p_k + self.n_k_min, min(p_k + self.n_k_max, self._state_space.num_factors) + 1)
        elif self.n_k_sample_mode == "bounded_below":
            n_k = np.random.randint(max(p_k, self.n_k_min), self.n_k_max + 1)
        elif self.n_k_sample_mode == "random":
            n_k = np.random.randint(self.n_k_min, self.n_k_max + 1)
        else:
            raise KeyError(f"Unknown mode: {self.n_k_sample_mode=}")
        # we're done!
        return p_k, n_k

    def _sample_shared_indices(self, p_k, n_k):
        p_shared_indices = np.random.choice(
            self._state_space.num_factors, size=self._state_space.num_factors - p_k, replace=False
        )
        # sample for negative
        if self.n_k_is_shared:
            n_shared_indices = p_shared_indices[: self._state_space.num_factors - n_k]
        else:
            n_shared_indices = np.random.choice(
                self._state_space.num_factors, size=self._state_space.num_factors - n_k, replace=False
            )
        # we're done!
        return p_shared_indices, n_shared_indices

    def _resample_factors(self, anchor_factors):
        # sample positive
        positive_factors = sample_radius(
            anchor_factors,
            low=0,
            high=self._state_space.factor_sizes,
            r_low=self.p_radius_min,
            r_high=self.p_radius_max + 1,
        )
        # negative arguments
        if self.n_radius_sample_mode == "offset":
            sampled_radius = np.abs(anchor_factors - positive_factors)
            n_r_low = sampled_radius + self.n_radius_min
            n_r_high = sampled_radius + self.n_radius_max + 1
        elif self.n_radius_sample_mode == "bounded_below":
            sampled_radius = np.abs(anchor_factors - positive_factors)
            n_r_low = np.maximum(sampled_radius, self.n_radius_min)
            n_r_high = self.n_radius_max + 1
        elif self.n_radius_sample_mode == "random":
            n_r_low = self.n_radius_min
            n_r_high = self.n_radius_max + 1
        else:
            raise KeyError(f"Unknown mode: {self.n_radius_sample_mode=}")
        # sample negative
        negative_factors = sample_radius(
            anchor_factors, low=0, high=self._state_space.factor_sizes, r_low=n_r_low, r_high=n_r_high
        )
        # we're done!
        return positive_factors, negative_factors

    def _swap_factors(self, anchor_factors, positive_factors, negative_factors):
        if self._swap_metric == "k":
            p_dist = np.sum(anchor_factors == positive_factors)
            n_dist = np.sum(anchor_factors == negative_factors)
        elif self._swap_metric == "manhattan":
            p_dist = np.sum(np.abs(anchor_factors - positive_factors))
            n_dist = np.sum(np.abs(anchor_factors - negative_factors))
        elif self._swap_metric == "manhattan_norm":
            p_dist = np.sum(
                np.abs((anchor_factors - positive_factors) / np.subtract(self._state_space.factor_sizes, 1))
            )
            n_dist = np.sum(
                np.abs((anchor_factors - negative_factors) / np.subtract(self._state_space.factor_sizes, 1))
            )
        elif self._swap_metric == "euclidean":
            p_dist = np.linalg.norm(anchor_factors - positive_factors)
            n_dist = np.linalg.norm(anchor_factors - negative_factors)
        elif self._swap_metric == "euclidean_norm":
            p_dist = np.linalg.norm(
                (anchor_factors - positive_factors) / np.subtract(self._state_space.factor_sizes, 1)
            )
            n_dist = np.linalg.norm(
                (anchor_factors - negative_factors) / np.subtract(self._state_space.factor_sizes, 1)
            )
        else:
            raise KeyError
        # perform swap
        if n_dist < p_dist:
            positive_factors, negative_factors = negative_factors, positive_factors
            # log.warning(f'Swapped factors based on metric: {self._swap_metric}')
        # return factors
        return positive_factors, negative_factors

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # END CLASS                                                             #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #


# ========================================================================= #
# HELPER                                                                    #
# ========================================================================= #


class FactorSizeError(Exception):
    pass


def normalise_range(mins, maxs, sizes):
    sizes = np.array(sizes)
    # compute the bounds for each factor
    mins = np.broadcast_to(mins, sizes.shape).copy()
    maxs = np.broadcast_to(maxs, sizes.shape).copy()
    mins[mins < 0] += sizes[mins < 0] + 1
    maxs[maxs < 0] += sizes[maxs < 0] + 1
    # check that min <= max
    assert np.all(mins <= maxs)
    # check that everything is in the right range [1, -1]
    assert np.all(0 <= mins) and np.all(mins <= sizes)
    assert np.all(0 <= maxs) and np.all(maxs <= sizes)
    # return merged
    return mins, maxs


def normalise_range_pair(min_max: Union[int, Tuple[int, int]], sizes):
    min_max = np.array(min_max)
    # if not a 2 tuple, repeat. This fixes the min == max.
    if min_max.shape == ():
        min_max = min_max.repeat(2)
    # check final shape
    assert min_max.shape == (2,)
    # get values
    return normalise_range(*min_max, sizes)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #

# if __name__ == '__main__':

# def conf(data, ci=0.95, two_tailed=True):
#     import scipy.stats
#     ci = ((1 - ci) / 2) if two_tailed else (1 - ci)
#     return scipy.stats.t.ppf(1 - ci, N - 1) * np.std(data) / np.sqrt(N)
#
# def conf_interval(data, ci=0.95):
#     mean = np.mean(data)
#     c = conf(data, ci, two_tailed=True)
#     return np.array([mean - c, mean + c])
#
# dataset = GroundTruthDatasetTriples(
#     XYGridData(),
#     # factor sampling
#     p_k_range=(1, 1),
#     n_k_range=(1, 1),
#     n_k_sample_mode='random',
#     n_k_is_shared=True,
#     # radius sampling
#     p_radius_range=(1, 1),
#     n_radius_range=(1, -1),
#     n_radius_sample_mode='offset',
#     # final checks
#     swap_metric=None,
# )
#
# print(dataset.data.factor_sizes)
# print()
#
# stats = defaultdict(list)
# N = min(len(dataset), 10_000)
#
# for i in tqdm(range(N)):
#     # CHECK DIFFERENT FACTORS
#     p_k, n_k = dataset._sample_num_factors()
#     stats['p_k'].append(p_k)
#     stats['n_k'].append(n_k)
#     # CHECK ALL SAMPLING
#     a, p, n = dataset.sample_factors(i)
#     # print(f'a:{a} p:{p} n:{n} | ap:{np.abs(a - p)} an:{np.abs(a - n)} | pn:{np.abs(p - n)}')
#     stats['a_p_diffs'].append(np.sum(a != p))
#     stats['a_n_diffs'].append(np.sum(a != n))
#     stats['a_p_ave_dist'].append(np.abs(a - p).sum())
#     stats['a_n_ave_dist'].append(np.abs(a - n).sum())
#     stats['p_n_ave_dist'].append(np.abs(p - n).sum())
#
# print(dataset.data.factor_sizes)
# for k, v in stats.items():
#     print(f'{k}: {np.around(np.mean(v), 3)} ± {np.around(conf(v, 0.95), 3)}')
#
# # check that resample radius is working correctly!
# dataset = GroundTruthDatasetTriples(
#     XYMultiGridData(1, 4),
#     # factor sampling
#     p_k_range=(1, 1),
#     n_k_range=(1, 1),
#     n_k_sample_mode='random',
#     n_k_is_shared=True,
#     # radius sampling
#     p_radius_range=(1, 1),
#     n_radius_range=(1, 1),
#     n_radius_sample_mode='offset',
#     # final checks
#     swap_metric=None,
# )
#
# for pair in dataset:
#     obs0, _, obs1 = np.array(pair[0], dtype='int'), np.array(pair[1], dtype='int'), np.array(pair[2], dtype='int')
#     # CHECKS
#     diff = np.abs(obs1 - obs0)
#     diff_coords = np.array(np.where(diff > 0)).T
#     assert len(diff_coords) == 2  # check max changes
#     dist = np.abs(diff_coords[0] - diff_coords[1])
#     assert np.sum(dist > 0) == 1  # check max changes
#     assert np.max(dist) == 2      # check radius
#     # INFO
#     print(concat_lines(*[((obs > 0) * [1, 2, 4]).sum(axis=-1) for obs in (obs0, obs1)]), '\n')

# import time
# t = time.time_ns()
# N = 10000
# samples = []
# for i in range(N):
#     sample = sample_radius(3, 0, 7, 0, 4)
#     samples.append(sample)
# samples = np.array(samples)
#
# u, c = np.unique(samples, return_counts=True)
# idxs = np.argsort(u)
# print(u[idxs])
# print(c[idxs] / N)
# ts = (time.time_ns() - t) / 1000_000_000
# print(ts * 1000, 'ms')
# print(N / ts, 'per/s')



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import warnings
from functools import lru_cache
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np
import scipy.stats
import torch
from PIL import Image

from disent.util import to_numpy

log = logging.getLogger(__name__)


# ========================================================================= #
# operations on Images/Animations                                           #
# ========================================================================= #


# get bg color -- TODO: add support for more
_BG_COLOR_DTYPE_MAP = {
    "uint8": 127,
    np.uint8: 127,
    torch.uint8: 127,
    np.dtype("uint8"): 127,
    "float16": 0.5,
    np.float16: 0.5,
    torch.float16: 0.5,
    np.dtype("float16"): 0.5,
    "float32": 0.5,
    np.float32: 0.5,
    torch.float32: 0.5,
    np.dtype("float32"): 0.5,
    "float64": 0.5,
    np.float64: 0.5,
    torch.float64: 0.5,
    np.dtype("float64"): 0.5,
}


def make_image_grid(
    images: Sequence[np.ndarray], pad: int = 8, border: bool = True, bg_color=None, num_cols: Optional[int] = None
):
    """
    Convert a list of images into a single image that is a grid of those images.
    :param images: list of input images, all the same size: (I, H, W, C) or (I, H, W)
    :param pad: the number of pixels between images
    :param border: if there should be a border around the grid
    :param bg_color: the background color to use for padding (can be a float, int or RGB tuple)
    :param num_cols: the number of output columns in the grid. None for auto square, -1 for rows==1, or > 0 for that many cols.
    :return: single output image:  (H', W') or (H', W', C)
    """
    # first, second, third channels are the (H, W, C)
    # get image sizes
    img_shape, ndim = np.array(images[0].shape), images[0].ndim
    assert ndim == 2 or ndim == 3, f"images have wrong number of channels: {img_shape}"
    assert np.all(img_shape == img.shape for img in images), "Images are not the same shape!"
    # get image size and channels
    img_size = img_shape[:2]
    if ndim == 3:
        assert img_shape[2] in (1, 3, 4), f"Invalid number of channels for an image: {img_shape}"
    # get bg color
    if bg_color is None:
        bg_color = _BG_COLOR_DTYPE_MAP[images[0].dtype]
    # grid sizes
    num_rows, num_cols = _get_grid_size(len(images), num_cols=num_cols)
    grid_size = (img_size + pad) * [num_rows, num_cols] + (pad if border else -pad)
    # image sizes including padding on one side
    deltas = img_size + pad
    offset = pad if border else 0
    # make image
    grid = np.full_like(images, fill_value=bg_color, shape=(*grid_size, *img_shape[2:]))
    # fill image
    for i, img in enumerate(images):
        y0, x0 = offset + deltas * [i // num_cols, i % num_cols]
        y1, x1 = img_size + [y0, x0]
        grid[y0:y1, x0:x1, ...] = img
    return grid


def make_animated_image_grid(
    list_of_animated_images: Sequence[np.ndarray],
    pad: int = 8,
    border: bool = True,
    bg_color=None,
    num_cols: Optional[int] = None,
):
    """
    :param list_of_animated_images: list of input images, with the second dimension the number of frames: : (I, F, H, W, C) or (I, F, H, W)
    :param pad: the number of pixels between images
    :param border: if there should be a border around the grid
    :param bg_color: the background color to use for padding (can be a float, int or RGB tuple)
    :param num_cols: the number of output columns in the grid. None for auto square, -1 for rows==1, or > 0 for that many cols.
    :return: animated output image: (F, H', W') or (F, H', W', C)
    """
    # first channel is the image (I)
    # second channel is the frame (F)
    # third, fourth, fifth channels are the (H, W, C)
    # -- (I, F, H, W, C)
    frames = []
    for list_of_images in zip(*list_of_animated_images):
        frames.append(make_image_grid(list_of_images, pad=pad, border=border, bg_color=bg_color, num_cols=num_cols))
    return to_numpy(frames)


# ========================================================================= #
# Calculations/Heuristics                                                   #
# ========================================================================= #


def _get_grid_size(n: int, num_cols: Optional[int] = None):
    """
    Determine the number of rows and columns, given the total number of elements n.
    - if num_cols is None:     rows x cols is as square as possible
    - if num_cols is a number: minimum number of rows needed is returned.
    - if num_cols <= 0:        only 1 row is returned
    :return: (num_rows, num_cols)
    """
    if num_cols is None:
        num_cols = int(np.ceil(n**0.5))
    elif num_cols <= 0:
        num_cols = n
    num_rows = (n + num_cols - 1) // num_cols
    return num_rows, num_cols


# ========================================================================= #
# Factor Cycle Generators                                                   #
# ========================================================================= #


def _get_interval_factor_traversal(factor_size: int, num_frames: int, start_index: int = 0):
    """
    Cycles through the state space in a single cycle.
    eg. num_indices=5, num_frames=7 returns: [0,1,1,2,3,3,4]
    eg. num_indices=4, num_frames=7 returns: [0,0,1,2,2,2,3]  # TODO: this result is weird
    """
    grid = np.linspace(0, factor_size - 1, num=num_frames, endpoint=True)
    grid = np.int64(np.around(grid))
    grid = (start_index + grid) % factor_size
    return grid


def _get_cycle_factor_traversal(factor_size: int, num_frames: int, start_index: int = 0):
    """
    Cycles through the state space in a single cycle.
    eg. num_indices=5, num_frames=7 returns: [0,1,3,4,3,2,1]
    eg. num_indices=4, num_frames=7 returns: [0,1,2,3,2,2,0]
    """
    assert start_index == 0, f"cycle factor traversal mode only supports start_index==0, got: {repr(start_index)}"
    grid = _get_interval_factor_traversal(factor_size=factor_size, num_frames=num_frames, start_index=0)
    grid = np.concatenate([grid[0::2], grid[1::2][::-1]])
    return grid


def _get_cycle_factor_traversal_from_start(factor_size: int, num_frames: int, start_index: int = 0, ends: bool = False):
    all_idxs = np.array(
        [
            *range(start_index, factor_size - (1 if ends else 0)),
            *reversed(range(0, factor_size)),
            *range(1 if ends else 0, start_index),
        ]
    )
    selected_idxs = _get_interval_factor_traversal(factor_size=len(all_idxs), num_frames=num_frames, start_index=0)
    grid = all_idxs[selected_idxs]
    # checks
    assert all_idxs[0] == start_index, "Please report this bug!"
    assert grid[0] == start_index, "Please report this bug!"
    return grid


def _get_cycle_factor_traversal_from_start_ends(
    factor_size: int, num_frames: int, start_index: int = 0, ends: bool = True
):
    return _get_cycle_factor_traversal_from_start(
        factor_size=factor_size, num_frames=num_frames, start_index=start_index, ends=ends
    )


_FACTOR_TRAVERSALS = {
    "interval": _get_interval_factor_traversal,
    "cycle": _get_cycle_factor_traversal,
    "cycle_from_start": _get_cycle_factor_traversal_from_start,
    "cycle_from_start_ends": _get_cycle_factor_traversal_from_start_ends,
}


def get_idx_traversal(factor_size: int, num_frames: int, mode: str = "interval", start_index: int = 0):
    try:
        traversal_fn = _FACTOR_TRAVERSALS[mode]
    except KeyError:
        raise KeyError(f"Invalid factor traversal mode: {repr(mode)}")
    return traversal_fn(factor_size=factor_size, num_frames=num_frames, start_index=start_index)


# ========================================================================= #
# Cycle Generators | FROM: disentanglement_lib                              #
# ========================================================================= #


def cycle_gaussian(starting_value, num_frames, loc=0.0, scale=1.0):
    """
    Cycles through the quantiles of a Gaussian in a single cycle.
    # ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~ #
    Copyright 2018 The DisentanglementLib Authors. All rights reserved.
    Licensed under the Apache License, Version 2.0
    https://github.com/google-research/disentanglement_lib
    # ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~ #
    """
    starting_prob = scipy.stats.norm.cdf(starting_value, loc=loc, scale=scale)
    grid = np.linspace(starting_prob, starting_prob + 2.0, num=num_frames, endpoint=False)
    grid -= np.maximum(0, 2 * grid - 2)
    grid += np.maximum(0, -2 * grid)
    grid = np.minimum(grid, 0.999)
    grid = np.maximum(grid, 0.001)
    return np.array([scipy.stats.norm.ppf(i, loc=loc, scale=scale) for i in grid])


def cycle_interval(starting_value, num_frames, min_val, max_val):
    """
    Cycles through the state space in a single cycle.
    # ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~ #
    Copyright 2018 The DisentanglementLib Authors. All rights reserved.
    Licensed under the Apache License, Version 2.0
    https://github.com/google-research/disentanglement_lib
    # ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~ #
    """
    starting_in_01 = (starting_value - min_val) / (max_val - min_val)
    starting_in_01 = np.nan_to_num(starting_in_01)  # handle division by zero, prints warning
    grid = np.linspace(starting_in_01, starting_in_01 + 2.0, num=num_frames, endpoint=False)
    grid -= np.maximum(0, 2 * grid - 2)
    grid += np.maximum(0, -2 * grid)
    return grid * (max_val - min_val) + min_val


# ========================================================================= #
# END                                                                      #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from typing import Dict
from typing import NoReturn
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np

from disent.dataset.data import NumpyFileGroundTruthData
from disent.dataset.util.datafile import DataFile
from disent.dataset.util.datafile import DataFileHashed
from disent.dataset.util.datafile import DataFileHashedDl
from disent.util.inout.paths import modify_name_keep_ext

log = logging.getLogger(__name__)


# ========================================================================= #
# Binary Matrix Helper Functions                                            #
# - https://cs.nyu.edu/~ylclab/data/norb-v1.0-small                         #
# ========================================================================= #


_BINARY_MATRIX_TYPES = {
    0x1E3D4C55: "uint8",  # byte matrix
    0x1E3D4C54: "int32",  # integer matrix
    0x1E3D4C56: "int16",  # short matrix
    0x1E3D4C51: "float32",  # single precision matrix
    0x1E3D4C53: "float64",  # double precision matrix
    # 0x1E3D4C52: '???',    # packed matrix -- not sure what this is?
}


def read_binary_matrix_bytes(bytes):
    """
    Read the binary matrix data
    - modified from disentanglement_lib

    Binary Matrix File Format Specification
        * The Header:
            - dtype:      4 bytes
            - ndim:       4 bytes, little endian
            - dim_sizes: (4 * min(3, ndim)) bytes
        * Handling the number of dimensions:
            - If there are less than 3 dimensions, then dim[1] and dim[2] are both: 1
            - Elif there are 3 or more dimensions, then the header will contain further size information.
        * Handling Matrix Data:
            - Little endian matrix data comes after the header,
              the index of the last dimension changes the fastest.
    """
    # header: dtype, ndim, dim_sizes
    dtype = int(np.frombuffer(bytes, dtype="int32", count=1, offset=0))  # bytes [0, 4)
    ndim = int(np.frombuffer(bytes, dtype="int32", count=1, offset=4))  # bytes [4, 8)
    stored_ndim = max(3, ndim)  # stores minimum of 3 dimensions even for 1D array
    dims = np.frombuffer(bytes, dtype="int32", count=stored_ndim, offset=8)[0:ndim]  # bytes [8, 8 + eff_dim * 4)
    # matrix: data
    data = np.frombuffer(bytes, dtype=_BINARY_MATRIX_TYPES[dtype], count=-1, offset=8 + stored_ndim * 4)
    data = data.reshape(tuple(dims))
    # done
    return data


def read_binary_matrix_file(file, gzipped: bool = True):
    import gzip

    # this does not seem to copy the bytes, which saves memory
    with (gzip.open if gzipped else open)(file, "rb") as f:
        return read_binary_matrix_bytes(bytes=f.read())


# ========================================================================= #
# Norb Functions                                                            #
# ========================================================================= #


def read_norb_dataset(
    dat_path: str, cat_path: str, info_path: str, gzipped=True, sort=True, add_channel_dim: bool = True
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Load The Normalised Dataset
    * dat:
        - images (5 categories, 5 instances, 6 lightings, 9 elevations, and 18 azimuths)
          + shape: (N, H, W, 1)
    * cat:
        - initial ground truth factor:
            0. category of images (0 for animal, 1 for human, 2 for plane, 3 for truck, 4 for car).
    * info:
        - additional ground truth factors:
            1. the instance in the category (0 to 9)
            2. the elevation (0 to 8, which mean cameras are 30, 35,40,45,50,55,60,65,70 degrees from the horizontal respectively)
            3. the azimuth (0,2,4,...,34, multiply by 10 to get the azimuth in degrees)
            4. the lighting condition (0 to 5)
    """
    # read the dataset
    dat = read_binary_matrix_file(dat_path, gzipped=gzipped)
    cat = read_binary_matrix_file(cat_path, gzipped=gzipped)
    info = read_binary_matrix_file(info_path, gzipped=gzipped)
    # collect the ground truth factors
    factors = np.column_stack([cat, info])  # append info to categories
    factors[:, 3] = factors[:, 3] / 2  # azimuth values are even numbers, convert to indices
    images = dat[:, 0]  # images are in pairs, only use the first. TODO: what is the second of each?
    # order the images and factors
    if sort:
        indices = np.lexsort(factors[:, [4, 3, 2, 1, 0]].T)
        images = images[indices]
        factors = factors[indices]
    # add the channel dimension
    if add_channel_dim:
        images = images[:, :, :, None]
        assert images.ndim == 4
    else:
        assert images.ndim == 3
    # done!
    return images, factors


def resave_norb_archive(
    in_dat_path: str, in_cat_path: str, in_info_path: str, new_save_file: str, in_gzipped=True, overwrite: bool = False
):
    # load the array
    images, factors = read_norb_dataset(
        dat_path=in_dat_path,
        cat_path=in_cat_path,
        info_path=in_info_path,
        gzipped=in_gzipped,
        sort=True,
        add_channel_dim=True,
    )
    assert images.shape == (24300, 96, 96, 1)
    # save the array
    from disent.dataset.util.formats.npz import save_dataset_array

    save_dataset_array(images, new_save_file, overwrite=overwrite, save_key="images")


def resave_norb_resized(orig_converted_file: str, new_resized_file: str, overwrite=False, size: int = 64):
    # load the array
    norb_array = np.load(orig_converted_file)["images"]
    assert norb_array.shape == (24300, 96, 96, 1)
    # save the array
    from disent.dataset.util.formats.npz import save_resized_dataset_array

    save_resized_dataset_array(norb_array, new_resized_file, overwrite=overwrite, size=size, save_key="images")


# ========================================================================= #
# Data Files                                                                #
# ========================================================================= #


class DataFileSmallNorb(DataFileHashed):
    """
    download the smallnorb dataset and convert it to a numpy file.
    """

    def __init__(
        self,
        datafile_dat: DataFile,
        datafile_cat: DataFile,
        datafile_info: DataFile,
        out_name: str,
        out_hash: Optional[Union[str, Dict[str, str]]],
        hash_type: str = "md5",
        hash_mode: str = "fast",
    ):
        self._datafile_dat = datafile_dat
        self._datafile_cat = datafile_cat
        self._datafile_info = datafile_info
        # initialize
        super().__init__(file_name=out_name, file_hash=out_hash, hash_type=hash_type, hash_mode=hash_mode)

    def _prepare(self, out_dir: str, out_file: str) -> NoReturn:
        resave_norb_archive(
            in_dat_path=self._datafile_dat.prepare(out_dir),
            in_cat_path=self._datafile_cat.prepare(out_dir),
            in_info_path=self._datafile_info.prepare(out_dir),
            new_save_file=out_file,
            in_gzipped=True,
            overwrite=True,
        )


class DataFileSmallNorbResized(DataFileHashed):
    def __init__(
        self,
        norb_datafile: DataFileSmallNorb,
        # - convert file name
        out_hash: Optional[Union[str, Dict[str, str]]],
        out_name: Optional[str] = None,
        out_size: int = 64,
        # - hash settings
        hash_type: str = "md5",
        hash_mode: str = "fast",
    ):
        self._out_size = out_size
        self._norb_datafile = norb_datafile
        super().__init__(
            file_name=modify_name_keep_ext(self._norb_datafile.out_name, suffix=f"_x{out_size}")
            if (out_name is None)
            else out_name,
            file_hash=out_hash,
            hash_type=hash_type,
            hash_mode=hash_mode,
        )

    def _prepare(self, out_dir: str, out_file: str):
        log.debug("Preparing Orig SmallNorb Data:")
        norb_path = self._norb_datafile.prepare(out_dir)
        log.debug("Generating Resized SmallNorb Data:")
        resave_norb_resized(
            orig_converted_file=norb_path, new_resized_file=out_file, overwrite=True, size=self._out_size
        )


# ========================================================================= #
# dataset_norb                                                              #
# ========================================================================= #


class SmallNorbData(NumpyFileGroundTruthData):
    """
    Small NORB Dataset
    - https://cs.nyu.edu/~ylclab/data/norb-v1.0-small/

    # reference implementation: https://github.com/google-research/disentanglement_lib/blob/master/disentanglement_lib/data/ground_truth/norb.py
    # TODO: add ability to randomly sample the instance so that this corresponds to disentanglement_lib
    """

    name = "smallnorb"

    factor_names = ("category", "instance", "elevation", "rotation", "lighting")
    factor_sizes = (5, 5, 9, 18, 6)  # TOTAL: 24300
    img_shape = (96, 96, 1)

    DATA_FILE_TRAIN = DataFileSmallNorb(
        datafile_dat=DataFileHashedDl(
            uri="https://cs.nyu.edu/~ylclab/data/norb-v1.0-small/smallnorb-5x46789x9x18x6x2x96x96-training-dat.mat.gz",
            uri_hash={"fast": "92560cccc7bcbd6512805e435448b62d", "full": "66054832f9accfe74a0f4c36a75bc0a2"},
        ),
        datafile_cat=DataFileHashedDl(
            uri="https://cs.nyu.edu/~ylclab/data/norb-v1.0-small/smallnorb-5x46789x9x18x6x2x96x96-training-cat.mat.gz",
            uri_hash={"fast": "348fc3ccefd651d69f500611988b5dcd", "full": "23c8b86101fbf0904a000b43d3ed2fd9"},
        ),
        datafile_info=DataFileHashedDl(
            uri="https://cs.nyu.edu/~ylclab/data/norb-v1.0-small/smallnorb-5x46789x9x18x6x2x96x96-training-info.mat.gz",
            uri_hash={"fast": "f1b170c16925867c05f58608eb33ba7f", "full": "51dee1210a742582ff607dfd94e332e3"},
        ),
        out_name="smallnorb_train.npz",
        out_hash={"fast": "a2c7de23c57b16c71b79dc2c884ecd67", "full": "7dabafbfafa0eb9b0115452f82d1491e"},
        hash_mode="fast",
    )

    DATA_FILE_TEST = DataFileSmallNorb(
        datafile_dat=DataFileHashedDl(
            uri="https://cs.nyu.edu/~ylclab/data/norb-v1.0-small/smallnorb-5x01235x9x18x6x2x96x96-testing-dat.mat.gz",
            uri_hash={"fast": "9aee0b474a4fc2a2ec392b463efb8858", "full": "e4ad715691ed5a3a5f138751a4ceb071"},
        ),
        datafile_cat=DataFileHashedDl(
            uri="https://cs.nyu.edu/~ylclab/data/norb-v1.0-small/smallnorb-5x01235x9x18x6x2x96x96-testing-cat.mat.gz",
            uri_hash={"fast": "8cfae0679f5fa2df7a0aedfce90e5673", "full": "5aa791cd7e6016cf957ce9bdb93b8603"},
        ),
        datafile_info=DataFileHashedDl(
            uri="https://cs.nyu.edu/~ylclab/data/norb-v1.0-small/smallnorb-5x01235x9x18x6x2x96x96-testing-info.mat.gz",
            uri_hash={"fast": "d2703a3f95e7b9a970ad52e91f0aaf6a", "full": "a9454f3864d7fd4bb3ea7fc3eb84924e"},
        ),
        out_name="smallnorb_test.npz",
        out_hash={"fast": "ff027c01e14faea9a0d427d641d3bd8e", "full": "cec16d74a1b075fe4117b5167e16ceff"},
        hash_mode="fast",
    )

    # override
    data_key = "images"

    def __init__(self, data_root: Optional[str] = None, prepare: bool = False, is_test: bool = False, transform=None):
        self._is_test = is_test
        # initialize
        super().__init__(data_root=data_root, prepare=prepare, transform=transform)

    @property
    def datafile(self) -> DataFile:
        return self.DATA_FILE_TEST if self._is_test else self.DATA_FILE_TRAIN


class SmallNorb64Data(SmallNorbData):
    img_shape = (64, 64, 1)

    DATA_FILE_TRAIN = DataFileSmallNorbResized(
        norb_datafile=SmallNorbData.DATA_FILE_TRAIN,
        out_name="smallnorb_train_x64.npz",
        out_size=64,
        out_hash={"fast": "74a3c02ea5a649313ea245a3fe271d3b", "full": "88ce361b2198ee577e60da2be9daa0e8"},
        hash_mode="fast",
    )

    DATA_FILE_TEST = DataFileSmallNorbResized(
        norb_datafile=SmallNorbData.DATA_FILE_TEST,
        out_name="smallnorb_test_x64.npz",
        out_size=64,
        out_hash={"fast": "37bf364479c0954ecd707ace349541ef", "full": "6bfd93eb6454d9d24dba13cac5f1ef3e"},
        hash_mode="fast",
    )


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


if __name__ == "__main__":
    import torch
    from tqdm import tqdm

    from disent.dataset.transform import ToImgTensorF32

    logging.basicConfig(level=logging.DEBUG)

    for is_test in [False, True]:
        # original dataset
        data_96 = SmallNorbData(prepare=True, is_test=is_test, transform=ToImgTensorF32(size=64))
        for i in tqdm(data_96, desc="norb_x96 -> 64"):
            pass
        # resized dataset
        data_64 = SmallNorb64Data(prepare=True, is_test=is_test, transform=ToImgTensorF32(size=64))
        for i in tqdm(data_64, desc="norb_x64"):
            pass
        # check equivalence
        for obs_96, obs_64 in tqdm(zip(data_96, data_64), desc="equivalence"):
            assert torch.allclose(obs_96, obs_64)



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import numpy as np
import torch

# ========================================================================= #
# helper                                                                    #
# ========================================================================= #


def torch_normalize(tensor: torch.Tensor, dims=None, dtype=None):
    # get min & max values
    if dims is not None:
        m, M = tensor, tensor
        for dim in dims:
            m, M = m.min(dim=dim, keepdim=True).values, M.max(dim=dim, keepdim=True).values
    else:
        m, M = tensor.min(), tensor.max()
    # scale tensor
    return (tensor.to(dtype=dtype) - m) / (M - m)  # automatically converts to float32 if needed


# ========================================================================= #
# polyfill - in later versions of pytorch                                   #
# ========================================================================= #


def torch_nan_to_num(input, nan=0.0, posinf=None, neginf=None):
    output = input.clone()
    if nan is not None:
        output[torch.isnan(input)] = nan
    if posinf is not None:
        output[input == np.inf] = posinf
    if neginf is not None:
        output[input == -np.inf] = neginf
    return output


# ========================================================================= #
# Torch Dim Helper                                                          #
# ========================================================================= #


def torch_unsqueeze_l(input: torch.Tensor, n: int):
    """
    Add n new axis to the left.

    eg. a tensor with shape (2, 3) passed to this function
        with n=2 will input in an output shape of (1, 1, 2, 3)
    """
    assert n >= 0, f"number of new axis cannot be less than zero, given: {repr(n)}"
    return input[((None,) * n) + (...,)]


def torch_unsqueeze_r(input: torch.Tensor, n: int):
    """
    Add n new axis to the right.

    eg. a tensor with shape (2, 3) passed to this function
        with n=2 will input in an output shape of (2, 3, 1, 1)
    """
    assert n >= 0, f"number of new axis cannot be less than zero, given: {repr(n)}"
    return input[(...,) + ((None,) * n)]


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from functools import lru_cache
from typing import Tuple
from typing import Union

import numpy as np
import torch
import torch.nn.functional as F

# ========================================================================= #
# helper functions for moving dimensions                                    #
# ========================================================================= #


@lru_cache(maxsize=32)
def _get_2d_reshape_info(shape: Tuple[int, ...], dims: Union[int, Tuple[int, ...]] = -1):
    if isinstance(dims, int):
        dims = (dims,)
    # number of dimensions & remove negatives
    ndim = len(shape)
    dims = tuple((ndim + d) if d < 0 else d for d in dims)
    # check that we have at least 2 dims & that all values are valid
    assert all(0 <= d < ndim for d in dims)
    # return new shape
    if ndim == 1:
        return [0], (1, *shape)
    # check resulting shape
    assert ndim >= 2
    # get dims
    dims_X = set(dims)
    dims_B = set(range(ndim)) - dims_X
    # sort dims
    dims_X = sorted(dims_X)
    dims_B = sorted(dims_B)
    # compute shape
    shape = np.array(shape)
    size_B = int(np.prod(shape[dims_B]))
    size_X = int(np.prod(shape[dims_X]))
    # variables
    moved_end_dims = tuple(dims_X[::-1])
    reshape_size = (size_B, size_X)
    # sort dims
    return moved_end_dims, reshape_size


def torch_dims_at_end_2d(tensor: torch.Tensor, dims: Union[int, Tuple[int, ...]] = -1, return_undo_data=True):
    # get dim info
    moved_end_dims, reshape_size = _get_2d_reshape_info(tensor.shape, dims=dims)
    # move all axes
    for d in moved_end_dims:
        tensor = torch.moveaxis(tensor, d, -1)
    moved_shape = tensor.shape
    # reshape
    tensor = torch.reshape(tensor, reshape_size)
    # return all info
    if return_undo_data:
        return tensor, moved_shape, moved_end_dims
    else:
        return tensor


def torch_undo_dims_at_end_2d(tensor: torch.Tensor, moved_shape, moved_end_dims):
    # reshape
    tensor = torch.reshape(tensor, moved_shape)
    # undo moving of dims
    for d in moved_end_dims[::-1]:
        tensor = torch.moveaxis(tensor, -1, d)
    # reshape
    return tensor


# ========================================================================= #
# Improved differentiable torchsort functions                               #
# ========================================================================= #


def torch_soft_sort(
    tensor: torch.Tensor,
    dims: Union[int, Tuple[int, ...]] = -1,
    regularization="l2",
    regularization_strength=1.0,
    leave_dims_at_end=False,
):
    # we import it locally so that we don't have to install this
    import torchsort

    # reorder the dimensions
    tensor, moved_shape, moved_end_dims = torch_dims_at_end_2d(tensor, dims=dims, return_undo_data=True)
    # sort the last dimension of the 2D tensors
    tensor = torchsort.soft_sort(tensor, regularization=regularization, regularization_strength=regularization_strength)
    # undo the reorder operation
    if leave_dims_at_end:
        return tensor
    return torch_undo_dims_at_end_2d(tensor, moved_shape=moved_shape, moved_end_dims=moved_end_dims)


def torch_soft_rank(
    tensor: torch.Tensor,
    dims: Union[int, Tuple[int, ...]] = -1,
    regularization="l2",
    regularization_strength=1.0,
    leave_dims_at_end=False,
):
    # we import it locally so that we don't have to install this
    import torchsort

    # reorder the dimensions
    tensor, moved_shape, moved_end_dims = torch_dims_at_end_2d(tensor, dims=dims, return_undo_data=True)
    # sort the last dimension of the 2D tensors
    tensor = torchsort.soft_rank(tensor, regularization=regularization, regularization_strength=regularization_strength)
    # undo the reorder operation
    if leave_dims_at_end:
        return tensor
    return torch_undo_dims_at_end_2d(tensor, moved_shape=moved_shape, moved_end_dims=moved_end_dims)


# ========================================================================= #
# Differentiable Spearman rank correlation                                  #
# ========================================================================= #


def multi_spearman_rank_loss(
    pred: torch.Tensor,
    targ: torch.Tensor,
    dims: Union[int, Tuple[int, ...]] = -1,
    reduction="mean",
    regularization="l2",
    regularization_strength=1.0,
    nan_to_num=False,
) -> torch.Tensor:
    """
    Like spearman_rank_loss but computes the loss over the specified
    dims, taking the average over the remaining unspecified dims.

    eg. given a tensor of shape (B, C, H, W) and specifying dims=(1, 3), the
        loss is computed over the C & W dims, by internally moving the dimensions
        to the end (B, H, C, W) and the reshaping the array to (B*H, C*W) before
        passing the array to spearman_rank_loss and returning the mean loss.
    """
    assert pred.shape == targ.shape
    pred = torch_dims_at_end_2d(pred, dims=dims, return_undo_data=False)
    targ = torch_dims_at_end_2d(targ, dims=dims, return_undo_data=False)
    # compute
    assert reduction == "mean", 'only supports reduction="mean"'
    return spearman_rank_loss(
        pred=pred,
        targ=targ,
        reduction=reduction,
        regularization=regularization,
        regularization_strength=regularization_strength,
        nan_to_num=nan_to_num,
    )


def spearman_rank_loss(
    pred: torch.Tensor,
    targ: torch.Tensor,
    reduction="mean",
    regularization="l2",
    regularization_strength=1.0,
    nan_to_num=False,
):
    """
    Compute the spearman correlation coefficient.
    - Supports 1D and 2D tensors.
    - For a 2D tensor of shape (B, N), the spearman correlation
      coefficient is computed over the last axis, with
      the mean taken over the B resulting values.
    - If the tensor is 1D of shape (N,) it is converted
      automatically to a 2D tensor of shape (1, N)

    Adapted from: https://github.com/teddykoker/torchsort
    """
    # TODO: I think something is wrong with this...
    # we import it locally so that we don't have to install this
    import torchsort

    # add missing dim
    if pred.ndim == 1:
        pred, targ = pred.reshape(1, -1), targ.reshape(1, -1)
    assert pred.shape == targ.shape
    assert pred.ndim == 2
    # sort the last dimension of the 2D tensors
    assert (
        regularization == "l2"
    ), "Only l2 regularization is currently supported for torchsort, others can result in memory leaks. See the torchsort github page for the bug report."
    pred = torchsort.soft_rank(pred, regularization=regularization, regularization_strength=regularization_strength)
    targ = torchsort.soft_rank(targ, regularization=regularization, regularization_strength=regularization_strength)
    # compute individual losses
    # TODO: this can result in nan values, what to do then?
    pred = pred - pred.mean(dim=-1, keepdim=True)
    pred = pred / pred.norm(dim=-1, keepdim=True)
    targ = targ - targ.mean(dim=-1, keepdim=True)
    targ = targ / targ.norm(dim=-1, keepdim=True)
    # replace nan values
    if nan_to_num:
        pred = torch.nan_to_num(pred, nan=0.0)
        targ = torch.nan_to_num(targ, nan=0.0)
    # compute the final loss
    loss = (pred * targ).sum(dim=-1)
    # reduce the loss
    if reduction == "mean":
        return loss.mean()
    elif reduction == "none":
        return loss
    else:
        raise KeyError(f"Invalid reduction mode: {repr(reduction)}")


# ========================================================================= #
# loss functions                                                            #
# ========================================================================= #


def torch_mse_rank_loss(pred, targ, dims=-1, regularization="l2", regularization_strength=1.0, reduction="mean"):
    return F.mse_loss(
        torch_soft_rank(
            pred,
            dims=dims,
            regularization=regularization,
            regularization_strength=regularization_strength,
            leave_dims_at_end=False,
        ),
        torch_soft_rank(
            targ,
            dims=dims,
            regularization=regularization,
            regularization_strength=regularization_strength,
            leave_dims_at_end=False,
        ),
        reduction=reduction,
    )


# ========================================================================= #
# end                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import pickle
from dataclasses import asdict

import lightning as L
import pytest
import torch
from lightning.pytorch.callbacks import ModelCheckpoint
from torch.utils.data import DataLoader

from disent.dataset import DisentDataset
from disent.dataset.data import XYObjectData
from disent.dataset.sampling import GroundTruthPairSampler
from disent.dataset.sampling import GroundTruthSingleSampler
from disent.dataset.sampling import GroundTruthTripleSampler
from disent.dataset.sampling import RandomSampler
from disent.dataset.transform import ToImgTensorF32
from disent.frameworks.ae import *
from disent.frameworks.vae import *
from disent.model import AutoEncoder
from disent.model.ae import DecoderLinear
from disent.model.ae import EncoderLinear
from disent.util.function import wrapped_partial
from disent.util.seeds import TempNumpySeed
from disent.util.seeds import seed
from docs.examples.extend_experiment.code.weaklysupervised__si_adavae import SwappedInputAdaVae
from docs.examples.extend_experiment.code.weaklysupervised__si_betavae import SwappedInputBetaVae

# ========================================================================= #
# TEST FRAMEWORKS                                                           #
# ========================================================================= #


_TEST_FRAMEWORKS = [
    # AE - unsupervised
    (Ae, dict(), XYObjectData),
    # AE - unsupervised - EXP
    (DataOverlapTripletAe, dict(overlap_mine_triplet_mode="hard_neg"), XYObjectData),
    # AE - weakly supervised
    # <n/a>
    # AE - weakly supervised - EXP
    (AdaAe, dict(), XYObjectData),
    # AE - supervised
    (TripletAe, dict(), XYObjectData),
    # AE - supervised - EXP
    (AdaNegTripletAe, dict(), XYObjectData),
    # VAE - unsupervised
    (Vae, dict(), XYObjectData),
    (BetaVae, dict(), XYObjectData),
    (DipVae, dict(), XYObjectData),
    (DipVae, dict(dip_mode="i"), XYObjectData),
    (InfoVae, dict(), XYObjectData),
    (DfcVae, dict(), XYObjectData),
    (DfcVae, dict(), wrapped_partial(XYObjectData, rgb=False)),
    (BetaTcVae, dict(), XYObjectData),
    # VAE - unsupervised - EXP
    (DataOverlapTripletVae, dict(overlap_mine_triplet_mode="none"), XYObjectData),
    (DataOverlapTripletVae, dict(overlap_mine_triplet_mode="semi_hard_neg"), XYObjectData),
    (DataOverlapTripletVae, dict(overlap_mine_triplet_mode="hard_neg"), XYObjectData),
    (DataOverlapTripletVae, dict(overlap_mine_triplet_mode="hard_pos"), XYObjectData),
    (DataOverlapTripletVae, dict(overlap_mine_triplet_mode="easy_pos"), XYObjectData),
    # VAE - weakly supervised
    (AdaVae, dict(), XYObjectData),
    (AdaVae, dict(ada_average_mode="ml-vae"), XYObjectData),
    (AdaGVaeMinimal, dict(), XYObjectData),
    # VAE - weakly supervised - EXP
    (SwappedInputAdaVae, dict(swap_chance=1.0), XYObjectData),
    (SwappedInputBetaVae, dict(swap_chance=1.0), XYObjectData),
    # VAE - supervised
    (TripletVae, dict(), XYObjectData),
    (TripletVae, dict(detach_decoder=True, disable_reg_loss=True), XYObjectData),
]


@pytest.mark.parametrize(["Framework", "cfg_kwargs", "Data"], _TEST_FRAMEWORKS)
def test_frameworks(Framework, cfg_kwargs, Data):
    DataSampler = {
        1: GroundTruthSingleSampler,
        2: GroundTruthPairSampler,
        3: GroundTruthTripleSampler,
    }[Framework.REQUIRED_OBS]

    data = XYObjectData() if (Data is None) else Data()
    dataset = DisentDataset(data, DataSampler(), transform=ToImgTensorF32())
    dataloader = DataLoader(dataset=dataset, batch_size=4, shuffle=True, num_workers=0)

    framework = Framework(
        model=AutoEncoder(
            encoder=EncoderLinear(x_shape=data.x_shape, z_size=6, z_multiplier=2 if issubclass(Framework, Vae) else 1),
            decoder=DecoderLinear(x_shape=data.x_shape, z_size=6),
        ),
        cfg=Framework.cfg(**cfg_kwargs),
    )

    # test pickling before training
    pickle.dumps(framework)

    # train!
    trainer = L.Trainer(logger=False, enable_checkpointing=False, max_steps=256, fast_dev_run=True)
    trainer.fit(framework, dataloader)

    # test pickling after training, something may have changed!
    pickle.dumps(framework)


@pytest.mark.parametrize(["Framework", "cfg_kwargs", "Data"], _TEST_FRAMEWORKS)
def test_framework_checkpointing(Framework, cfg_kwargs, Data, tmp_path):
    DataSampler = {
        1: GroundTruthSingleSampler,
        2: GroundTruthPairSampler,
        3: GroundTruthTripleSampler,
    }[Framework.REQUIRED_OBS]

    data = XYObjectData() if (Data is None) else Data()
    dataset = DisentDataset(data, DataSampler(), transform=ToImgTensorF32())
    dataloader = DataLoader(dataset=dataset, batch_size=4, shuffle=True, num_workers=0)

    framework = Framework(
        model=AutoEncoder(
            encoder=EncoderLinear(x_shape=data.x_shape, z_size=6, z_multiplier=2 if issubclass(Framework, Vae) else 1),
            decoder=DecoderLinear(x_shape=data.x_shape, z_size=6),
        ),
        cfg=Framework.cfg(**cfg_kwargs),
    )

    # save checkpoints during fit()
    cpk = ModelCheckpoint(dirpath=tmp_path, save_last=True)

    # train!
    trainer = L.Trainer(
        default_root_dir=tmp_path, enable_checkpointing=True, callbacks=[cpk], max_steps=2, fast_dev_run=False
    )

    framework.hparams.update(cfg_kwargs)
    if trainer.loggers:
        for logger in trainer.loggers:
            logger.log_hyperparams(framework.hparams)

    trainer.fit(framework, dataloader)

    # load checkpoint
    checkpoints = list(tmp_path.glob("*.ckpt"))
    assert len(checkpoints) == 2
    for c in checkpoints:
        framework_checkpoint = Framework.load_from_checkpoint(c)
        assert isinstance(framework_checkpoint, Framework)


@pytest.mark.parametrize(["Framework", "cfg_kwargs", "Data"], _TEST_FRAMEWORKS)
def test_framework_pickling(Framework, cfg_kwargs, Data):
    framework = Framework(
        model=AutoEncoder(
            encoder=EncoderLinear(x_shape=(64, 64, 3), z_size=6, z_multiplier=2 if issubclass(Framework, Vae) else 1),
            decoder=DecoderLinear(x_shape=(64, 64, 3), z_size=6),
        ),
        cfg=Framework.cfg(**cfg_kwargs),
    )
    # test pickling!
    pickle.dumps(framework)


def test_framework_config_defaults():
    # import torch
    # we test that defaults are working recursively
    assert asdict(BetaVae.cfg()) == dict(
        optimizer="adam",
        optimizer_kwargs=None,
        recon_loss="mse",
        disable_aug_loss=False,
        detach_decoder=False,
        disable_rec_loss=False,
        disable_reg_loss=False,
        loss_reduction="mean",
        latent_distribution="normal",
        kl_loss_mode="direct",
        beta=0.003,
    )
    assert asdict(BetaVae.cfg(recon_loss="bce", kl_loss_mode="approx")) == dict(
        optimizer="adam",
        optimizer_kwargs=None,
        recon_loss="bce",
        disable_aug_loss=False,
        detach_decoder=False,
        disable_rec_loss=False,
        disable_reg_loss=False,
        loss_reduction="mean",
        latent_distribution="normal",
        kl_loss_mode="approx",
        beta=0.003,
    )


def test_ada_vae_similarity():
    seed(42)

    data = XYObjectData()
    dataset = DisentDataset(data, sampler=RandomSampler(num_samples=2), transform=ToImgTensorF32())
    dataloader = DataLoader(dataset, batch_size=3, num_workers=0)

    model = AutoEncoder(
        encoder=EncoderLinear(x_shape=data.x_shape, z_size=25, z_multiplier=2),
        decoder=DecoderLinear(x_shape=data.x_shape, z_size=25, z_multiplier=1),
    )

    adavae0 = AdaGVaeMinimal(model=model, cfg=AdaGVaeMinimal.cfg())
    adavae1 = AdaVae(model=model, cfg=AdaVae.cfg())
    adavae2 = AdaVae(
        model=model,
        cfg=AdaVae.cfg(
            ada_average_mode="gvae",
            ada_thresh_mode="symmetric_kl",
            ada_thresh_ratio=0.5,
        ),
    )

    batch = next(iter(dataloader))

    # TODO: add a TempNumpySeed equivalent for torch
    seed(777)
    result0a = adavae0.do_training_step(batch, 0)
    seed(777)
    result0b = adavae0.do_training_step(batch, 0)
    assert torch.allclose(result0a, result0b), f"{result0a} does not match {result0b}"

    seed(777)
    result1a = adavae1.do_training_step(batch, 0)
    seed(777)
    result1b = adavae1.do_training_step(batch, 0)
    assert torch.allclose(result1a, result1b), f"{result1a} does not match {result1b}"

    seed(777)
    result2a = adavae2.do_training_step(batch, 0)
    seed(777)
    result2b = adavae2.do_training_step(batch, 0)
    assert torch.allclose(result2a, result2b), f"{result2a} does not match {result2b}"

    # check similar
    assert torch.allclose(result0a, result1a), f"{result0a} does not match {result1a}"
    assert torch.allclose(result1a, result2a), f"{result1a} does not match {result2a}"


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import Tuple
from typing import final

# ========================================================================= #
# Base Sampler                                                              #
# ========================================================================= #


class BaseDisentSampler(object):
    def uninit_copy(self) -> "BaseDisentSampler":
        raise NotImplementedError

    def __init__(self, num_samples: int):
        self._num_samples = num_samples

    @property
    def num_samples(self) -> int:
        return self._num_samples

    __initialized = False

    @final
    def init(self, dataset) -> "BaseDisentSampler":
        if self.__initialized:
            raise RuntimeError(
                f"Sampler: {repr(self.__class__.__name__)} has already been initialized, are you sure it is not being reused?"
            )
        # initialize
        self.__initialized = True
        self._init(dataset)
        return self

    def _init(self, dataset):
        """
        you can override this method to initialize the class!
        """
        pass

    @property
    def is_init(self) -> bool:
        return self.__initialized

    def _sample_idx(self, idx: int) -> Tuple[int, ...]:
        raise NotImplementedError

    def sample(self, idx: int) -> Tuple[int, ...]:
        # check that we have been initialized!
        if not self.is_init:
            raise RuntimeError(f"{self.__class__.__name__} has not been initialized! call `sampler.init(gt_data)`")
        # sample values
        idxs = self._sample_idx(idx)
        # check values
        if len(idxs) != self.num_samples:
            raise RuntimeError(
                f"{self.__class__.__name__} returned incorrect number of samples, required: {self.num_samples}, got: {len(idxs)}"
            )
        # return values
        return idxs

    def __call__(self, idx: int) -> Tuple[int, ...]:
        return self.sample(idx)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os
from abc import ABCMeta
from typing import Any
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np
from torch.utils.data import Dataset

from disent.dataset.data._raw import Hdf5Dataset
from disent.dataset.util.datafile import DataFile
from disent.dataset.util.datafile import DataFileHashedDlH5
from disent.dataset.util.state_space import StateSpace
from disent.util.inout.paths import ensure_dir_exists

log = logging.getLogger(__name__)


# ========================================================================= #
# ground truth data                                                         #
# ========================================================================= #


class GroundTruthData(Dataset, StateSpace):
    """
    Dataset that corresponds to some state space or ground truth factors
    """

    def __init__(self, transform=None):
        self._transform = transform
        super().__init__(
            factor_sizes=self.factor_sizes,
            factor_names=self.factor_names,
        )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Overridable Defaults                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def name(self):
        name = self.__class__.__name__
        if name.endswith("Data"):
            name = name[: -len("Data")]
        return name.lower()

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # State Space                                                           #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def factor_names(self) -> Tuple[str, ...]:
        raise NotImplementedError()

    @property
    def factor_sizes(self) -> Tuple[int, ...]:
        raise NotImplementedError()

    def state_space_copy(self) -> StateSpace:
        """
        :return: Copy this ground truth dataset as a StateSpace, discarding everything else!
        """
        return StateSpace(
            factor_sizes=self.factor_sizes,
            factor_names=self.factor_names,
        )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Properties                                                            #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def x_shape(self) -> Tuple[int, ...]:
        # shape as would be for a single observation in a torch batch
        # eg. C x H x W
        H, W, C = self.img_shape
        return (C, H, W)

    @property
    def img_shape(self) -> Tuple[int, ...]:
        # shape as would be for an original image
        # eg. H x W x C
        raise NotImplementedError()

    @property
    def img_channels(self) -> int:
        channels = self.img_shape[-1]
        assert channels in (
            1,
            3,
        ), f"invalid number of channels for dataset: {self.__class__.__name__}, got: {repr(channels)}, required: 1 or 3"
        return channels

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Overrides                                                             #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def __getitem__(self, idx):
        obs = self._get_observation(idx)
        if self._transform is not None:
            obs = self._transform(obs)
        return obs

    def _get_observation(self, idx):
        raise NotImplementedError

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # EXTRAS                                                                #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def sample_random_obs_traversal(
        self, f_idx: int = None, base_factors=None, num: int = None, mode="interval", obs_collect_fn=None
    ) -> Tuple[np.ndarray, np.ndarray, Union[List[Any], Any]]:
        """
        Same API as sample_random_factor_traversal, but also
        returns the corresponding indices and uncollated list of observations
        """
        factors = self.sample_random_factor_traversal(f_idx=f_idx, base_factors=base_factors, num=num, mode=mode)
        indices = self.pos_to_idx(factors)
        obs = [self[i] for i in indices]
        if obs_collect_fn is not None:
            obs = obs_collect_fn(obs)
        return factors, indices, obs


# ========================================================================= #
# Basic Array Ground Truth Dataset                                          #
# ========================================================================= #


class ArrayGroundTruthData(GroundTruthData):
    def __init__(
        self,
        array,
        factor_names: Tuple[str, ...],
        factor_sizes: Tuple[int, ...],
        array_chn_is_last: bool = True,
        x_shape: Optional[Tuple[int, ...]] = None,
        transform=None,
    ):
        self.__factor_names = tuple(factor_names)
        self.__factor_sizes = tuple(factor_sizes)
        self._array = array
        # get shape
        if x_shape is not None:
            C, H, W = x_shape
        elif array_chn_is_last:
            H, W, C = array.shape[1:]
        else:
            C, H, W = array.shape[1:]
        # set observation shape
        self.__img_shape = (H, W, C)
        # initialize
        super().__init__(transform=transform)
        # check shapes -- it is up to the user to handle which method they choose
        assert (array.shape[1:] == self.img_shape) or (array.shape[1:] == self.x_shape)

    @property
    def array(self):
        return self._array

    @property
    def factor_names(self) -> Tuple[str, ...]:
        return self.__factor_names

    @property
    def factor_sizes(self) -> Tuple[int, ...]:
        return self.__factor_sizes

    @property
    def img_shape(self) -> Tuple[int, ...]:
        return self.__img_shape

    def _get_observation(self, idx):
        # TODO: INVESTIGATE! I think this implements a lock,
        #       hindering multi-threaded environments?
        return self._array[idx]

    @classmethod
    def new_like(cls, array, gt_data: GroundTruthData, array_chn_is_last: bool = True):
        # TODO: should this not copy the x_shape and transform?
        return cls(
            array=array,
            factor_names=gt_data.factor_names,
            factor_sizes=gt_data.factor_sizes,
            array_chn_is_last=array_chn_is_last,
            x_shape=None,  # infer from array
            transform=None,
        )


# ========================================================================= #
# disk ground truth data                                                    #
# TODO: data & datafile preparation should be split out from                #
#       GroundTruthData, instead GroundTruthData should be a wrapper        #
# ========================================================================= #


class _DiskDataMixin(object):
    # attr this class defines in _mixin_disk_init
    _data_dir: str

    def _mixin_disk_init(self, data_root: Optional[str] = None, prepare: bool = False):
        # get root data folder
        if data_root is None:
            data_root = self.default_data_root
        else:
            data_root = os.path.abspath(os.path.expanduser(data_root))
        # get class data folder
        self._data_dir = ensure_dir_exists(os.path.join(data_root, self.name))
        log.info(f"{self.name}: data_dir_share={repr(self._data_dir)}")
        # prepare everything
        if prepare:
            for datafile in self.datafiles:
                log.debug(f"[preparing]: {datafile} into data dir: {self._data_dir}")
                datafile.prepare(self.data_dir)

    @property
    def data_dir(self) -> str:
        return self._data_dir

    @property
    def default_data_root(self):
        return os.path.abspath(os.environ.get("DISENT_DATA_ROOT", "data/dataset"))

    @property
    def datafiles(self) -> Sequence[DataFile]:
        raise NotImplementedError

    @property
    def name(self) -> str:
        raise NotImplementedError


class DiskGroundTruthData(_DiskDataMixin, GroundTruthData, metaclass=ABCMeta):

    """
    Dataset that prepares a list DataObjects into some local directory.
    - This directory can be
    """

    def __init__(self, data_root: Optional[str] = None, prepare: bool = False, transform=None):
        super().__init__(transform=transform)
        # get root data folder
        self._mixin_disk_init(data_root=data_root, prepare=prepare)


class NumpyFileGroundTruthData(DiskGroundTruthData, metaclass=ABCMeta):
    """
    Dataset that loads a numpy file from a DataObject
    - if the dataset is contained in a key, set the `data_key` property
    """

    def __init__(self, data_root: Optional[str] = None, prepare: bool = False, transform=None):
        super().__init__(data_root=data_root, prepare=prepare, transform=transform)
        # load dataset
        load_path = os.path.join(self.data_dir, self.datafile.out_name)
        if load_path.endswith(".gz"):
            import gzip

            with gzip.GzipFile(load_path, "r") as load_file:
                self._data = np.load(load_file)
        else:
            self._data = np.load(load_path)
        # load from the key if specified
        if self.data_key is not None:
            self._data = self._data[self.data_key]

    def _get_observation(self, idx):
        return self._data[idx]

    @property
    def datafiles(self) -> Sequence[DataFile]:
        return [self.datafile]

    @property
    def datafile(self) -> DataFile:
        raise NotImplementedError

    @property
    def data_key(self) -> Optional[str]:
        # can override this!
        return None


class _Hdf5DataMixin(object):
    # attrs this class defines in _mixin_hdf5_init
    _in_memory: bool
    _attrs: dict
    _data: Union[Hdf5Dataset, np.ndarray]

    def _mixin_hdf5_init(self, h5_path: str, h5_dataset_name: str = "data", in_memory: bool = False):
        # variables
        self._in_memory = in_memory
        # load the h5py dataset
        data = Hdf5Dataset(
            h5_path=h5_path,
            h5_dataset_name=h5_dataset_name,
        )
        # load attributes
        self._attrs = data.get_attrs()
        # handle different memory modes
        if self._in_memory:
            # Load the entire dataset into memory if required
            # indexing dataset objects returns numpy array
            # instantiating np.array from the dataset requires double memory.
            self._data = data[:]
            self._data.flags.writeable = False
            data.close()
        else:
            # Load the dataset from the disk
            self._data = data

    def __len__(self):
        return len(self._data)

    @property
    def img_shape(self):
        shape = self._data.shape[1:]
        assert len(shape) == 3
        return shape

    # override from GroundTruthData
    def _get_observation(self, idx):
        return self._data[idx]


class Hdf5GroundTruthData(_Hdf5DataMixin, DiskGroundTruthData, metaclass=ABCMeta):
    """
    Dataset that loads an Hdf5 file from a DataObject
    - requires that the data object has the `out_dataset_name` attribute
      that points to the hdf5 dataset in the file to load.
    """

    def __init__(self, data_root: Optional[str] = None, prepare: bool = False, in_memory=False, transform=None):
        super().__init__(data_root=data_root, prepare=prepare, transform=transform)
        # initialize mixin
        self._mixin_hdf5_init(
            h5_path=os.path.join(self.data_dir, self.datafile.out_name),
            h5_dataset_name=self.datafile.dataset_name,
            in_memory=in_memory,
        )

    @property
    def datafiles(self) -> Sequence[DataFileHashedDlH5]:
        return [self.datafile]

    @property
    def datafile(self) -> DataFileHashedDlH5:
        raise NotImplementedError


class SelfContainedHdf5GroundTruthData(_Hdf5DataMixin, GroundTruthData):
    def __init__(self, h5_path: str, in_memory=False, transform=None):
        # initialize mixin
        self._mixin_hdf5_init(
            h5_path=h5_path,
            h5_dataset_name="data",
            in_memory=in_memory,
        )
        # load attrs
        self._attr_name = self._attrs["dataset_name"].decode("utf-8")
        self._attr_factor_names = tuple(name.decode("utf-8") for name in self._attrs["factor_names"])
        self._attr_factor_sizes = tuple(int(size) for size in self._attrs["factor_sizes"])
        # set size
        (B, H, W, C) = self._data.shape
        self._img_shape = (H, W, C)
        # initialize!
        super().__init__(transform=transform)

    @property
    def name(self) -> str:
        return self._attr_name

    @property
    def factor_names(self) -> Tuple[str, ...]:
        return self._attr_factor_names

    @property
    def factor_sizes(self) -> Tuple[int, ...]:
        return self._attr_factor_sizes

    @property
    def img_shape(self) -> Tuple[int, ...]:
        return self._img_shape


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from numbers import Number
from typing import Any
from typing import Dict
from typing import Optional

import numpy as np
import torch

# TODO: matplotlib is not in requirements
from matplotlib import pyplot as plt

from disent.dataset import DisentDataset
from disent.dataset.util.state_space import NonNormalisedFactorIdxs
from disent.util.seeds import TempNumpySeed
from disent.util.visualize.vis_util import make_animated_image_grid
from disent.util.visualize.vis_util import make_image_grid

log = logging.getLogger(__name__)


# ========================================================================= #
# vars                                                                      #
# ========================================================================= #


_TORCH_NORMAL_TYPES = {torch.float16, torch.float32, torch.float64}

# torch.complex32 exists in 1.10, but was disabled in 1.11, and planned to be added in 1.12 again
if torch.version.__version__.startswith("1.11."):
    _TORCH_COMPLEX_TYPES = {torch.complex64}
else:
    _TORCH_COMPLEX_TYPES = {torch.complex32, torch.complex64}


# ========================================================================= #
# Matplotlib Helper                                                         #
# ========================================================================= #


def plt_imshow(img, figsize=12, show=False, **kwargs):
    # check image shape
    assert img.ndim == 3
    assert img.shape[-1] in (1, 3, 4)
    # figure size -- fixed width, adjust height according to image
    if isinstance(figsize, (int, str, Number)):
        size = np.array(img.shape[:2][::-1])
        figsize = tuple(size / size[0] * figsize)
    # create plot
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize, **kwargs)
    plt_hide_axis(ax)
    ax.imshow(img)
    fig.tight_layout()
    # done!
    if show:
        plt.show()
    return fig, ax


def _hide(hide, cond):
    assert hide in {True, False, "all", "edges", "none"}
    return (hide is True) or (hide == "all") or (hide == "edges" and cond)


def plt_subplots(
    nrows: int = 1,
    ncols: int = 1,
    # custom
    title=None,
    titles=None,
    row_labels=None,
    col_labels=None,
    title_size: int = None,
    titles_size: int = None,
    label_size: int = None,
    hide_labels="edges",  # none, edges, all
    hide_axis="edges",  # none, edges, all
    # plt.subplots:
    sharex: str = False,
    sharey: str = False,
    subplot_kw=None,
    gridspec_kw=None,
    **fig_kw,
):
    assert isinstance(nrows, int)
    assert isinstance(ncols, int)
    # check titles
    if titles is not None:
        titles = np.array(titles)
        if titles.ndim == 1:
            titles = np.stack([titles] + ([[None] * ncols] * (nrows - 1)), axis=0)
        assert titles.ndim == 2, f"invalid titles shape, must have 2 dims: {titles.shape}"
    # get labels
    if (row_labels is None) or isinstance(row_labels, str):
        row_labels = [row_labels] * nrows
    if (col_labels is None) or isinstance(col_labels, str):
        col_labels = [col_labels] * ncols
    assert len(row_labels) == nrows, "row_labels and nrows mismatch"
    assert len(col_labels) == ncols, "row_labels and nrows mismatch"
    # check titles
    if titles is not None:
        assert len(titles) == nrows
        assert len(titles[0]) == ncols
    # create subplots
    fig, axs = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        sharex=sharex,
        sharey=sharey,
        squeeze=False,
        subplot_kw=subplot_kw,
        gridspec_kw=gridspec_kw,
        **fig_kw,
    )
    # generate
    for y in range(nrows):
        for x in range(ncols):
            ax = axs[y, x]
            plt_hide_axis(ax, hide_xaxis=_hide(hide_axis, y != nrows - 1), hide_yaxis=_hide(hide_axis, x != 0))
            # modify ax
            if not _hide(hide_labels, y != nrows - 1):
                ax.set_xlabel(col_labels[x], fontsize=label_size)
            if not _hide(hide_labels, x != 0):
                ax.set_ylabel(row_labels[y], fontsize=label_size)
            # set title
            if titles is not None:
                if titles[y][x] is not None:
                    ax.set_title(titles[y][x], fontsize=titles_size)
    # set title
    fig.suptitle(title, fontsize=title_size)
    # done!
    return fig, axs


def plt_subplots_imshow(
    grid,
    # custom:
    title=None,
    titles=None,
    row_labels=None,
    col_labels=None,
    title_size: int = None,
    titles_size: int = None,
    label_size: int = None,
    hide_labels="edges",  # none, edges, all
    hide_axis="all",  # none, edges, all
    # tight_layout:
    subplot_padding: Optional[float] = 1.08,
    # plt.subplots:
    sharex: str = False,
    sharey: str = False,
    subplot_kw=None,
    gridspec_kw=None,
    # imshow
    vmin: float = None,
    vmax: float = None,
    # extra
    show: bool = False,
    imshow_kwargs: dict = None,
    **fig_kw,
):
    # TODO: add automatic height & width
    fig, axs = plt_subplots(
        nrows=len(grid),
        ncols=len(grid[0]),
        # custom
        title=title,
        titles=titles,
        row_labels=row_labels,
        col_labels=col_labels,
        title_size=title_size,
        titles_size=titles_size,
        label_size=label_size,
        hide_labels=hide_labels,  # none, edges, all
        hide_axis=hide_axis,  # none, edges, all
        # plt.subplots:
        sharex=sharex,
        sharey=sharey,
        subplot_kw=subplot_kw,
        gridspec_kw=gridspec_kw,
        **fig_kw,
    )
    # show images
    for y, x in np.ndindex(axs.shape):
        axs[y, x].imshow(grid[y][x], vmin=vmin, vmax=vmax, **(imshow_kwargs if imshow_kwargs else {}))
    fig.tight_layout(**({} if (subplot_padding is None) else dict(pad=subplot_padding)))
    # done!
    if show:
        plt.show()
    return fig, axs


def plt_hide_axis(
    ax, hide_xaxis=True, hide_yaxis=True, hide_border=True, hide_axis_labels=False, hide_axis_ticks=True, hide_grid=True
):
    if hide_xaxis:
        if hide_axis_ticks:
            ax.set_xticks([])
            ax.set_xticklabels([])
        if hide_axis_labels:
            ax.xaxis.label.set_visible(False)
    if hide_yaxis:
        if hide_axis_ticks:
            ax.set_yticks([])
            ax.set_yticklabels([])
        if hide_axis_labels:
            ax.yaxis.label.set_visible(False)
    if hide_border:
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
        ax.spines["left"].set_visible(False)
    if hide_grid:
        ax.grid(False)
    return ax


# ========================================================================= #
# Dataset Visualisation / Traversals                                        #
# ========================================================================= #


def visualize_dataset_traversal(
    dataset: DisentDataset,
    # inputs
    factor_names: Optional[NonNormalisedFactorIdxs] = None,
    num_frames: int = 9,
    seed: int = 777,
    base_factors=None,
    traverse_mode="cycle",
    # images & animations
    pad: int = 4,
    border: bool = True,
    bg_color: Number = None,
    # augment
    augment_fn: callable = None,
    data_mode: str = "raw",
    # output
    output_wandb: bool = False,
):
    """
    Generic function that can return multiple parts of the dataset & factor traversal pipeline.
    - This only evaluates what is needed to compute the next components.
    - The returned grid, image and animation will always have 3 channels, RGB

    Tasks include:
        - factor_idxs
        - factors
        - grid
        - image
        - image_wandb
        - image_plt
        - animation
        - animation_wandb
    """

    # get factors from dataset
    factor_idxs = dataset.gt_data.normalise_factor_idxs(factor_names)

    # get factor traversals
    with TempNumpySeed(seed):
        factors = np.stack(
            [
                dataset.gt_data.sample_random_factor_traversal(
                    f_idx, base_factors=base_factors, num=num_frames, mode=traverse_mode
                )
                for f_idx in factor_idxs
            ],
            axis=0,
        )

    # retrieve and augment image grid
    grid = [dataset.dataset_batch_from_factors(f, mode=data_mode) for f in factors]
    if augment_fn is not None:
        grid = [augment_fn(batch) for batch in grid]
    grid = np.stack(grid, axis=0)

    # TODO: this is kinda hacky, maybe rather add a check?
    # TODO: can this be moved into the `output_wandb` if statement?
    # - animations glitch out if they do not have 3 channels
    assert grid.ndim == 5, f"invalid number of dimensions, must be 5, got: {grid.ndim}"
    if grid.shape[-1] == 1:
        grid = grid.repeat(3, axis=-1)
    assert grid.shape[-1] in (
        1,
        3,
    ), f"invalid number of channels, must be 1 or 3, got shape: {grid.shape}. Note that the dataset or augment if specified should output HWC images, not CHW images!"

    # generate visuals
    image = make_image_grid(
        np.concatenate(grid, axis=0), pad=pad, border=border, bg_color=bg_color, num_cols=num_frames
    )
    animation = make_animated_image_grid(
        np.stack(grid, axis=0), pad=pad, border=border, bg_color=bg_color, num_cols=None
    )

    # convert to wandb
    if output_wandb:
        import wandb

        wandb_image = wandb.Image(image)
        wandb_animation = wandb.Video(np.transpose(animation, [0, 3, 1, 2]), fps=4, format="mp4")
        return (
            wandb_image,
            wandb_animation,
        )

    # return values
    return (
        grid,  # (FACTORS, NUM_FRAMES, H, W, C)
        image,  # ([[H+PAD]*[FACTORS+1]], [[W+PAD]*[NUM_FRAMES+1]], C)
        animation,  # (NUM_FRAMES, [H & FACTORS], [W & FACTORS], C) -- size is auto-chosen
    )


# ========================================================================= #
# END                                                                       #
# ========================================================================= #
